#! /urs/bin/env python

"""Code for creating simulations of grism time series observations.


1. Make 2 temporary copies of the input yaml file.
   Copy 1 will contain all catalogs except the catalog with the TSO source
       and have the mode set to 'wfss'
   Copy 2 will contain only the TSO source catalog with mode set to 'wfss'

   NOTE THAT STEPS 2 AND 3 CAN BE COMBINED INTO A SINGLE CALL TO WFSS_SIMULATOR
2. Run the catalog_seed_generator on Copy 1. Result will be a padded countrate image
   containing only the non-TSO sources, which have constant brightness with time
   (dare we think about moving targets? - No, not now. Save for a later PR)
3. Run the disperser on the seed image with non-TSO sources. Result will be 2d
   dispersed countrate image

4. Run the catalog_seed_generator on Copy 2. Result will be a padded 2d countrate
   image containing only the TSO source
5. Read in transmission curve
6. Use batman to generate lightcurves from the transmission spectrum for each of
   a grid of wavelengths. (What grid? Dispersed image is 10A per res. element, so
   use that resolution?)
7. Run the disperser using the original, unaltered stellar spectrum. Set 'cache=True'.
   Result will be the dispersed countrate image of the TSO source for times when
   the lightcurve is 1.0 everywhere
8. Create an output array to hold the frame-by-frame seed image. Probably easier if this
   array holds just the signal accumulated in a given frame for that frame, rather than
   the cumulative signal in that frame. Note that there will
   most likely be file-splitting happening at this point...
9. Determine which frames of the exposure will take place with the unaltered stellar
   spectrum. This will be all frames where the associated lightcurve is 1.0 everywhere.
   These frames will be simply copies of the outputs from step 7 plus step 4. To get
   around the issue of file splitting, easier to just keep a list of frame index
   numbers which this situation applies to.
10.For each of the remaining frames, run the disperser with the appropriate lightcurve
   (after running interp1d to turn it into a function). The frame seed will be this
   output plus that from step 4
11.As you go, translate into a cumulative frame by frame seed image, and save to a
   fits file and reset the array variable as you get to the limit of each segment
   part.
12.Run the dark current prep step
13.Run the observation generator

"""


import copy
import os
import sys
import argparse
import yaml

from astropy.io import ascii, fits
import astropy.units as u
import batman
import numpy as np
from NIRCAM_Gsim.grism_seed_disperser import Grism_seed
import pkg_resources
import pysiaf
from scipy.interpolate import interp1d

from mirage import wfss_simulator
from mirage.catalogs import catalog_generator, spectra_from_catalog
from mirage.seed_image import catalog_seed_image
from mirage.dark import dark_prep
from mirage.ramp_generator import obs_generator
from mirage.reference_files import crds_tools
from mirage.utils import read_fits
from mirage.utils.constants import CATALOG_YAML_ENTRIES, MEAN_GAIN_VALUES
from mirage.utils.file_splitting import find_file_splits, SplitFileMetaData
from mirage.utils import utils, file_io, backgrounds
from mirage.yaml import yaml_update


class GrismTSO():
    def __init__(self, parameter_file, SED_file=None, SED_normalizing_catalog_column=None,
                 final_SED_file=None, save_dispersed_seed=True, source_stamps_file=None,
                 extrapolate_SED=True, override_dark=None, disp_seed_filename=None, orders=["+1", "+2"]):
        """
        Parameters
        ----------
        parameter_file : str
            Name of input yaml file

        SED_file : str
            Name of hdf5 file containing spectra of some or all sources in
            the simulated data

        SED_normalizing_catalog_column : str
            Name of the column in the ascii source catalogs to use when
            normalizing input spectra in the ``SED_file`` IF some or all
            of the spectra in that file contain normalized fluxes

        final_SED_file : str
            Name of hdf5 file produced by Mirage that contains spectra for
            all simulated sources. In the case where some sources have only
            broadband magnitudes but no entries in the ``SED_file``, Mirage
            will interpolate/extrapolate to create continuum spectra and add
            these to the spectra in the ``SED_file``

        save_dispersed_seed : bool
            Save the dispersed seed image to a fits file

        source_stamps_file : str

        extrapolate_SED : bool
            Allow the disperser software to extrapolate when creating
            continuum spectra from source catalog magnitudes. This is
            done if the wavelength range covered by the filter magnitudes
            does not cover the entire wavelength range of the grism+filter

        override_dark : str
            Name of file containing a Mirage dark current object appropriate
            for the simulation. If provided, the ``dark_prep`` step will be
            skipped. If ``None``, then a new dark current object will be
            constructed

        disp_seed_filename : str
            Name of the file to save the dispsersed seed image into

        orders : list
            List of spectral orders to create during dispersion. Default
            for NIRCam is ["+1", "+2"]
        """

        # Use the MIRAGE_DATA environment variable
        env_var = 'MIRAGE_DATA'
        self.datadir = os.environ.get(env_var)
        if self.datadir is None:
            raise ValueError(("WARNING: {} environment variable is not set."
                              "This must be set to the base directory"
                              "containing the darks, cosmic ray, PSF, etc"
                              "input files needed for the simulation."
                              "These files must be downloaded separately"
                              "from the Mirage package.".format(env_var)))
        self.modpath = pkg_resources.resource_filename('mirage', '')

        # Set the user-input parameters
        self.paramfile = parameter_file
        self.SED_file = SED_file
        self.SED_normalizing_catalog_column = SED_normalizing_catalog_column
        self.final_SED_file = final_SED_file
        self.save_dispersed_seed = save_dispersed_seed
        self.source_stamps_file = source_stamps_file
        self.extrapolate_SED = extrapolate_SED
        self.override_dark = override_dark
        self.disp_seed_filename = disp_seed_filename
        self.orders = orders
        self.fullframe_apertures = ["NRCA5_FULL", "NRCB5_FULL", "NIS_CEN"]

        # Make sure the right combination of parameter files and SED file
        # are given
        self.param_checks()

    def calculate_exposure_time(self):
        """Calculate the total exposure time of the observation being
        simulated. Include time for resets between integrations

        Returns
        -------
        exposure_time : float
            Exposure time for the total exposuure, including reset frames,
            in seconds
        """
        self.frametime = utils.calc_frame_time(self.instrument, self.aperture, self.seed_dimensions[1],
                                               self.seed_dimensions[0], self.namps)
        return self.frametime * self.total_frames

    def create(self):
        """MAIN FUNCTION"""

        # Get parameters necessary to create the TSO data
        orig_parameters = self.get_param_info()
        subarray_table = utils.read_subarray_definition_file(orig_parameters['Reffiles']['subarray_defs'])
        orig_parameters = utils.get_subarray_info(orig_parameters, subarray_table)
        orig_parameters = utils.read_pattern_check(orig_parameters)

        self.basename = os.path.join(orig_parameters['Output']['directory'],
                                     orig_parameters['Output']['file'][0:-5].split('/')[-1])

        # Determine file splitting information. First get some basic info
        # on the exposure
        self.numints = orig_parameters['Readout']['nint']
        self.numgroups = orig_parameters['Readout']['ngroup']
        self.numframes = orig_parameters['Readout']['nframe']
        self.numskips = orig_parameters['Readout']['nskip']
        self.namps = orig_parameters['Readout']['namp']
        self.numresets = orig_parameters['Readout']['resets_bet_ints']
        self.frames_per_group, self.frames_per_int, self.total_frames = utils.get_frame_count_info(self.numints,
                                                                                                   self.numgroups,
                                                                                                   self.numframes,
                                                                                                   self.numskips,
                                                                                                   self.numresets)

        # Get gain map for later unit conversion
        #gainfile = orig_parameters['Reffiles']['gain']
        #gain, gainheader = file_io.read_gain_file(gainfile)

        # Make 2 copies of the input parameter file, separating the TSO
        # source from the other sources
        self.split_param_file(orig_parameters)

        print('Splitting background and TSO source into multiple yaml files')
        print('Running background sources through catalog_seed_image')
        print('background param file is:', self.background_paramfile)

        # Run the catalog_seed_generator on the non-TSO (background) sources
        background_direct = catalog_seed_image.Catalog_seed()
        background_direct.paramfile = self.background_paramfile
        background_direct.make_seed()
        background_segmentation_map = background_direct.seed_segmap

        # Stellar spectrum hdf5 file will be required, so no need to create one here.
        # Create hdf5 file with spectra of all sources if requested
        self.final_SED_file = spectra_from_catalog.make_all_spectra(self.catalog_files,
                                                                    input_spectra_file=self.SED_file,
                                                                    extrapolate_SED=self.extrapolate_SED,
                                                                    output_filename=self.final_SED_file,
                                                                    normalizing_mag_column=self.SED_normalizing_catalog_column)

        bkgd_waves, bkgd_fluxes = backgrounds.nircam_background_spectrum(orig_parameters,
                                                                         self.detector, self.module)

        # Run the disperser on the background sources. Add the background
        # signal here as well
        print('\n\nDispersing background sources\n\n')

        background_done = False
        background_seed_files = [background_direct.ptsrc_seed_filename,
                                 background_direct.galaxy_seed_filename,
                                 background_direct.extended_seed_filename]
        for seed_file in background_seed_files:
            if seed_file is not None:
                print("Dispersing seed image:", seed_file)
                disp = self.run_disperser(seed_file, orders=self.orders,
                                          add_background=not background_done,
                                          background_waves=bkgd_waves,
                                          background_fluxes=bkgd_fluxes,
                                          finalize=True)
                if not background_done:
                    # Background is added at the first opportunity. At this
                    # point, create an array to hold the final combined
                    # dispersed background
                    background_done = True
                    background_dispersed = copy.deepcopy(disp.final)
                else:
                    background_dispersed += disp.final

        # Run the catalog_seed_generator on the TSO source
        tso_direct = catalog_seed_image.Catalog_seed()
        tso_direct.paramfile = self.tso_paramfile
        tso_direct.make_seed()
        tso_segmentation_map = tso_direct.seed_segmap
        outside_tso_source = tso_segmentation_map == 0
        tso_segmentation_map[outside_tso_source] = background_segmentation_map[outside_tso_source]

        # Dimensions are (y, x)
        self.seed_dimensions = tso_direct.nominal_dims

        # Read in the transmission spectrum that goes with the TSO source
        tso_params = utils.read_yaml(self.tso_paramfile)
        tso_catalog_file = tso_params['simSignals']['tso_grism_catalog']

        tso_catalog = ascii.read(tso_catalog_file)

        transmission_file = tso_catalog['Transmission_spectrum'].data
        transmission_spectrum = ascii.read(transmission_file[0])

        # Calculate the total exposure time, including resets, to check
        # against the times provided in the catalog file.
        total_exposure_time = self.calculate_exposure_time() * u.second

        # Check to be sure the start and end times provided in the catalog
        # are enough to cover the length of the exposure.
        tso_catalog = self.tso_catalog_check(tso_catalog, total_exposure_time)

        # Use batman to create lightcurves from the transmission spectrum
        lightcurves, times = self.make_lightcurves(tso_catalog, self.frametime, transmission_spectrum)

        # Determine which frames of the exposure will take place with the unaltered stellar
        # spectrum. This will be all frames where the associated lightcurve is 1.0 everywhere.
        transit_frames, unaltered_frames = self.find_transit_frames(lightcurves)
        print('Frame numbers containing the transit: {} - {}'.format(np.min(transit_frames), np.max(transit_frames)))

        # Run the disperser using the original, unaltered stellar spectrum. Set 'cache=True'
        print('\n\nDispersing TSO source\n\n')
        grism_seed_object = self.run_disperser(tso_direct.seed_file, orders=self.orders,
                                               add_background=False, cache=True, finalize=True)

        # Crop dispersed seed images to correct final subarray size
        #no_transit_signal = grism_seed_object.final
        no_transit_signal = utils.crop_to_subarray(grism_seed_object.final, tso_direct.subarray_bounds)
        background_dispersed = utils.crop_to_subarray(background_dispersed, tso_direct.subarray_bounds)

        # Mulitp[ly the dispersed seed images by the flat field
        no_transit_signal *= tso_direct.flatfield
        background_dispersed *= tso_direct.flatfield

        # Save the dispersed seed images if requested
        if self.save_dispersed_seed:
            h_back = fits.PrimaryHDU(background_dispersed)
            h_back.header['EXTNAME'] = 'BACKGROUND_SOURCES'
            h_tso = fits.ImageHDU(grism_seed_object.final)
            h_tso.header['EXTNAME'] = 'TSO_SOURCE'
            hlist = fits.HDUList([h_back, h_tso])
            disp_filename = '{}_dispersed_seed_images.fits'.format(self.basename)
            hlist.writeto(disp_filename, overwrite=True)
            print('\nDispersed seed images (background sources and TSO source) saved to {}.\n\n'
                  .format(disp_filename))

        # Calculate file splitting info
        self.file_splitting()

        # Prepare for creating output files
        segment_file_dir = orig_parameters['Output']['directory']
        if orig_parameters['Readout']['pupil'][0].upper() == 'F':
            usefilt = 'pupil'
        else:
            usefilt = 'filter'
        segment_file_base = orig_parameters['Output']['file'].replace('.fits', '_')
        segment_file_base = '{}_{}_'.format(segment_file_base, orig_parameters['Readout'][usefilt])
        segment_file_base = os.path.join(segment_file_dir, segment_file_base)

        # Loop over frames and integrations up to the size of the segment
        # file.
        ints_per_segment = self.int_segment_indexes[1:] - self.int_segment_indexes[:-1]
        groups_per_segment = self.grp_segment_indexes[1:] - self.grp_segment_indexes[:-1]

        total_frame_counter = 0
        previous_segment = 1
        segment_part_number = 0
        segment_starting_int_number = 0

        self.segment_part_number = 0
        self.segment_ints = 0
        self.segment_frames = 0
        self.segment_frame_start_number = 0
        self.segment_int_start_number = 0
        self.part_int_start_number = 0
        self.part_frame_start_number = 0

        # Get split files' metadata
        split_meta = SplitFileMetaData(self.int_segment_indexes, self.grp_segment_indexes,
                                       self.file_segment_indexes, self.group_segment_indexes_g,
                                       self.frames_per_int, self.frames_per_group, self.frametime)

        # List of all output seed files
        self.seed_files = []

        counter = 0
        for i, int_dim in enumerate(ints_per_segment):
            int_start = self.int_segment_indexes[i]
            int_end = self.int_segment_indexes[i+1]
            for j, grp_dim in enumerate(groups_per_segment):
                initial_frame = self.grp_segment_indexes[j]
                # int_dim and grp_dim are the number of integrations and
                # groups in the current segment PART
                print("\n\nCurrent segment part contains: {} integrations and {} groups.".format(int_dim, grp_dim))
                print("Creating frame by frame dispersed signal")
                segment_seed = np.zeros((int_dim, grp_dim, self.seed_dimensions[0], self.seed_dimensions[1]))

                for integ in np.arange(int_dim):
                    overall_integration_number = int_start + integ
                    previous_frame = np.zeros(self.seed_dimensions)

                    for frame in np.arange(grp_dim):
                        #print('TOTAL FRAME COUNTER: ', total_frame_counter)
                        #print('integ and frame: ', integ, frame)
                        # If a frame is from the part of the lightcurve
                        # with no transit, then the signal in the frame
                        # comes from no_transit_signal
                        if total_frame_counter in unaltered_frames:
                            frame_only_signal = (background_dispersed + no_transit_signal) * self.frametime
                        # If the frame is from a part of the lightcurve
                        # where the transit is happening, then call the
                        # cached disperser with the appropriate lightcurve
                        elif total_frame_counter in transit_frames:
                            #print("{} is during the transit".format(total_frame_counter))
                            frame_transmission = lightcurves[total_frame_counter, :]
                            trans_interp = interp1d(transmission_spectrum['Wavelength'], frame_transmission)

                            for order in self.orders:
                                grism_seed_object.this_one[order].disperse_all_from_cache(trans_interp)
                            # Here is where we call finalize on the TSO object
                            # This will update grism_seed_object.final to
                            # contain the correct signal
                            grism_seed_object.finalize(Back=None, BackLevel=None)
                            cropped_grism_seed_object = utils.crop_to_subarray(grism_seed_object.final, tso_direct.subarray_bounds)
                            frame_only_signal = (background_dispersed + cropped_grism_seed_object) * self.frametime

                        # Now add the signal from this frame to that in the
                        # previous frame in order to arrive at the total
                        # cumulative signal
                        segment_seed[integ, frame, :, :] = previous_frame + frame_only_signal
                        previous_frame = copy.deepcopy(segment_seed[integ, frame, :, :])
                        total_frame_counter += 1

                    # At the end of each integration, increment the
                    # total_frame_counter by the number of resets between
                    # integrations
                    total_frame_counter += self.numresets

                # Use the split files' metadata
                self.segment_number = split_meta.segment_number[counter]
                self.segment_ints = split_meta.segment_ints[counter]
                self.segment_frames = split_meta.segment_frames[counter]
                self.segment_part_number = split_meta.segment_part_number[counter]
                self.segment_frame_start_number = split_meta.segment_frame_start_number[counter]
                self.segment_int_start_number = split_meta.segment_int_start_number[counter]
                self.part_int_start_number = split_meta.part_int_start_number[counter]
                self.part_frame_start_number = split_meta.part_frame_start_number[counter]
                counter += 1

                print('Overall integration number: ', overall_integration_number)
                segment_file_name = '{}seg{}_part{}_seed_image.fits'.format(segment_file_base,
                                                                            str(self.segment_number).zfill(3),
                                                                            str(self.segment_part_number).zfill(3))


                print('Segment int and frame start numbers: {} {}'.format(self.segment_int_start_number, self.segment_frame_start_number))
                #print('Part int and frame start numbers (ints and frames within the segment): {} {}'.format(self.part_int_start_number, self.part_frame_start_number))

                # Disperser output is always full frame. Crop to the
                # requested subarray if necessary
                if orig_parameters['Readout']['array_name'] not in self.fullframe_apertures:
                    print("Dispersed seed image size: {}".format(segment_seed.shape))
                    segment_seed = utils.crop_to_subarray(segment_seed, tso_direct.subarray_bounds)
                    #gain = utils.crop_to_subarray(gain, tso_direct.subarray_bounds)

                # Segmentation map will be centered in a frame that is larger
                # than full frame by a factor of sqrt(2), so crop appropriately
                print('Cropping segmentation map to appropriate aperture')
                segy, segx = tso_segmentation_map.shape
                dx = int((segx - tso_direct.nominal_dims[1]) / 2)
                dy = int((segy - tso_direct.nominal_dims[0]) / 2)
                segbounds = [tso_direct.subarray_bounds[0] + dx, tso_direct.subarray_bounds[1] + dy,
                             tso_direct.subarray_bounds[2] + dx, tso_direct.subarray_bounds[3] + dy]
                tso_segmentation_map = utils.crop_to_subarray(tso_segmentation_map, segbounds)

                # Convert seed image to ADU/sec to be consistent
                # with other simulator outputs
                gain = MEAN_GAIN_VALUES['nircam']['lw{}'.format(self.module.lower())]
                segment_seed /= gain

                # Update seed image header to reflect the
                # division by the gain
                tso_direct.seedinfo['units'] = 'ADU/sec'

                # Save the seed image. Save in units of ADU/sec
                print('Saving seed image')
                tso_seed_header = fits.getheader(tso_direct.seed_file)
                self.save_seed(segment_seed, tso_segmentation_map, tso_seed_header, orig_parameters) #,
                               #segment_number, segment_part_number)

        # Prepare dark current exposure if
        # needed.
        print('Running dark prep')
        d = dark_prep.DarkPrep()
        d.paramfile = self.paramfile
        d.prepare()

        # Combine into final observation
        print('Running observation generator')
        obs = obs_generator.Observation()
        obs.linDark = d.dark_files
        obs.seed = self.seed_files
        obs.segmap = tso_segmentation_map
        obs.seedheader = tso_direct.seedinfo
        obs.paramfile = self.paramfile
        obs.create()

    def file_splitting(self):
        """Determine file splitting details based on calculated data
        volume
        """
        frames_per_group = self.frames_per_int / self.numgroups
        self.split_seed, self.grp_segment_indexes, self.int_segment_indexes = find_file_splits(self.seed_dimensions[1],
                                                                                               self.seed_dimensions[0],
                                                                                               self.frames_per_int,
                                                                                               self.numints,
                                                                                               frames_per_group=frames_per_group)

        # If the file needs to be split, check to see what the splitting
        # would be in the case of groups rather than frames. This will
        # help align the split files between the seed image and the dark
        # object later (which is split by groups).
        if self.split_seed:
            split_seed_g, self.group_segment_indexes_g, self.file_segment_indexes = find_file_splits(self.seed_dimensions[1],
                                                                                                self.seed_dimensions[0],
                                                                                                self.numgroups,
                                                                                                self.numints)

            # In order to avoid the case of having a single integration
            # in the final file, which leads to rate rather than rateints
            # files in the pipeline, check to be sure that the integration
            # splitting indexes indicate the last split isn't a single
            # integration
            if len(self.file_segment_indexes) > 2:
                delta_int = self.file_segment_indexes[1:] - self.file_segment_indexes[0: -1]
                if delta_int[-1] == 1 and delta_int[0] != 1:
                    self.file_segment_indexes[-2] -= 1
                    print('Adjusted to avoid single integration: ', self.file_segment_indexes)

            # More adjustments related to segment numbers. We need to compare
            # the integration indexes for the seed images vs those for the final
            # data and make sure that there are no segments in the final data that
            # have no corresponding integrations from the seed images
            # Example: integration_segment_indexes = [0, 7, 8], and
            # self.file_segment_indexes = [0, 6, 8] - due to applying the adjustment
            # above. In this case as you step through integration_segment_indexes,
            # you see that (skipping 0), 7 and 8 both fall in the 6-8 bin in
            # self.file_segment_indexes. Nothing falls into the 0-7 bin, which
            # corresponds to segment 1. In this case, we need to adjust
            # integration_segment_indexes to make sure that all segments have data
            # associated with them.
            segnum_check = []
            for intnum in self.int_segment_indexes[1:]:
                segnum_check.append(np.where(intnum <= self.file_segment_indexes)[0][0])
            maxseg = max(segnum_check)
            for i in range(1, maxseg + 1):
                if i not in segnum_check:
                    self.int_segment_indexes = copy.deepcopy(self.file_segment_indexes)

        else:
            self.file_segment_indexes = np.array([0, self.numints])
            self.group_segment_indexes_g = np.array([0, self.numgroups])

        self.total_seed_segments = len(self.file_segment_indexes) - 1
        self.total_seed_segments_and_parts = (len(self.int_segment_indexes) - 1) * (len(self.grp_segment_indexes) - 1)

    @staticmethod
    def find_transit_frames(lightcurve_collection):
        """
        """
        no_transit = []
        transit = []
        for row in range(lightcurve_collection.shape[0]):
            lc = lightcurve_collection[row, :]
            if np.all(lc == 1.0):
                no_transit.append(row)
            else:
                transit.append(row)
        return transit, no_transit

    def get_param_info(self):
        """Collect needed information out of the parameter file. Check
        parameter values for correctness

        Returns
        -------
        parameters : dict
            Nested dictionary of parameters in the input yaml file
        """
        self.catalog_files = []
        parameters = utils.read_yaml(self.paramfile)

        # Create dictionary to use when looking in CRDS for reference files
        crds_dict = crds_tools.dict_from_yaml(parameters)

        # Expand reference file entries to be full path names
        parameters = utils.full_paths(parameters, self.modpath, crds_dict)

        try:
            CATALOG_YAML_ENTRIES.remove('tso_grism_catalog')
        except ValueError:
            pass

        try:
            CATALOG_YAML_ENTRIES.remove('tso_imaging_catalog')
        except ValueError:
            pass
        cats = [parameters['simSignals'][cattype] for cattype in CATALOG_YAML_ENTRIES]
        cats = [e for e in cats if e.lower() != 'none']
        self.catalog_files.extend(cats)

        if len(self.catalog_files) == 0:
            # If no background source catalogs are given, create a dummy point
            # source catalog and add it to the list. Without this, the
            # creation of the final SED file would fail. Put the source
            # down near the SEP and with a magnitude such that it won't
            # disturb anything
            filter_name = parameters['Readout']['filter'].lower()
            dummy_ptsrc = catalog_generator.PointSourceCatalog(ra=[0.], dec=[-89.])
            dummy_ptsrc.add_magnitude_column([40], instrument='nircam', filter_name=filter_name, magnitude_system='abmag')
            dummy_cat = 'dummy_ptsrc.cat'
            dummy_ptsrc.save(dummy_cat)
            self.catalog_files.append(dummy_cat)
            parameters['simSignals']['pointsource'] = dummy_cat

        self.instrument = parameters['Inst']['instrument'].lower()
        self.aperture = parameters['Readout']['array_name']
        try:
            self.namps = parameters['Readout']['namp']
        except KeyError:
            pass
        if self.instrument == 'niriss':
            self.module = None
            self.detetor = 'NIS'
        elif self.instrument == 'nircam':
            self.module = parameters['Readout']['array_name'][3]
            self.detector = parameters['Readout']['array_name'][0:5]
        else:
            raise ValueError("ERROR: Grism TSO mode not supported for {}".format(self.instrument))

        filter_name = parameters['Readout']['filter']
        pupil_name = parameters['Readout']['pupil']

        # In reality, the grism elements are in NIRCam's pupil wheel, and NIRISS's
        # filter wheel. But in the APT xml file, NIRISS grisms are in the pupil
        # wheel and the crossing filter is listed in the filter wheel. At that
        # point, NIRISS and NIRCam are consistent, so let's keep with this reversed
        # information
        if self.instrument == 'niriss':
            self.crossing_filter = pupil_name.upper()
            self.dispersion_direction = filter_name[-1].upper()
        elif self.instrument == 'nircam':
            if 'CLEAR' in pupil_name.upper():
                raise ValueError("ERROR: pupil cannot be CLEAR. It must be GRISMR or GRISMC.")
            self.crossing_filter = filter_name.upper()
            self.dispersion_direction = pupil_name[-1].upper()
        return parameters


    def make_lightcurves(self, catalog, frame_time, transmission_spec):
        """Given a transmission spectrum, create a series of lightcurves
        using ``batman``.

        Parameters
        ----------
        catalog : astropy.table.Table
            Table containing info from the TSO source catalog

        Returns
        -------
        lightcurves : numpy.ndarray
            2D array containing the light curve at each wavelengthin the
            transmission spectrum
        """
        params = batman.TransitParams()

        # planet radius (in units of stellar radii)
        # DUMMY VALUE FOR MODEL INSTANTIATION
        params.rp = 0.1

        params.a = catalog['Semimajor_axis_in_stellar_radii']  # semi-major axis (in units of stellar radii)
        params.inc = catalog['Orbital_inclination_deg']        # orbital inclination (in degrees)
        params.ecc = catalog['Eccentricity']                   # eccentricity
        params.w = catalog['Longitude_of_periastron']          # longitude of periastron (in degrees)
        params.limb_dark = catalog['Limb_darkening_model']     # limb darkening model

        # Limb darkening coefficients [u1, u2, u3, u4]
        params.u = [np.float(e) for e in catalog['Limb_darkening_coeffs'][0].split(',')]

        # Get the time units from the catalog
        time_units = u.Unit(catalog['Time_units'][0])
        start_time = catalog['Start_time'][0] * time_units
        end_time = catalog['End_time'][0] * time_units

        # Convert times to units of seconds to make working
        # with frametimes later easier
        start_time = start_time.to(u.second).value
        end_time = end_time.to(u.second).value
        params.t0 = (catalog['Time_of_inferior_conjunction'][0] * time_units).to(u.second).value  # time of inferior conjunction
        params.per = (catalog['Orbital_period'][0] * time_units).to(u.second).value       # orbital period

        # The time resolution must be one frametime since we will need one
        # lightcurve for each frame later
        time = np.arange(start_time, end_time, frame_time)  # times at which to calculate light curve
        model = batman.TransitModel(params, time)

        # Step along the transmission spectrum in wavelength space and
        # calculate a lightcurve at each step
        lightcurves = np.ones((len(time), len(transmission_spec['Transmission'])))
        for i, radius in enumerate(transmission_spec['Transmission']):
            params.rp = radius                          # updates planet radius
            new_flux = model.light_curve(params)        # recalculates light curve
            lightcurves[:, i] = new_flux

        # Save the 2D transmission data
        h0 = fits.PrimaryHDU(lightcurves)
        hdulist = fits.HDUList([h0])
        outfile = '{}{}'.format(self.basename, '_normalized_lightcurves_vs_time.fits')
        hdulist.writeto(outfile, overwrite=True)
        print('2D array of lightcurves vs time saved to: {}'.format(outfile))

        return lightcurves, time

    def param_checks(self):
        """Check validity of inputs
        """
        if self.orders not in [["+1"], ["+2"], ["+1", "+2"], None]:
            raise ValueError(("ERROR: Orders to be dispersed must be either None or some combination "
                              "of '+1', '+2'"))

    def run_disperser(self, direct_file, orders=["+1", "+2"], add_background=True,
                      background_waves=None, background_fluxes=None, cache=False, finalize=False):
        """Run the disperser on the given direct seed image.

        Parameters
        ----------
        direct_file : str
            Name of file containing direct seed image

        orders : list
            Orders to include in the dispersed image.

        add_background : bool
            If True, a 2D dispersed background image is created and
            added

        background_waves : numpy.ndarray
            1D array of wavelengths (in mocrons) to be used when creating
            the background image

        background_fluxes : numpy.ndarray
            1D array of fluxes (in units of cgs f_lambda) to be used
            when creating the background image

        cache : bool
            Whether or not to cache the dispersed object. If it is cached,
            then it can be used later

        finalize : bool
            If True, call the finalize function of the disperser in order to
            create the final dispersed image of the object

        Returns
        -------
        disp_seed : numpy.ndarray
            2D array containing the dispersed seed image
        """
        # Location of the configuration files needed for dispersion
        loc = os.path.join(self.datadir, "{}/GRISM_{}/".format(self.instrument,
                                                               self.instrument.upper()))

        # Determine the name of the background file to use, as well as the
        # orders to disperse.
        dmode = 'mod{}_{}'.format(self.module, self.dispersion_direction)
        orders = self.orders

        # Create dispersed seed image from the direct images
        disp_seed = Grism_seed([direct_file], self.crossing_filter,
                               dmode, config_path=loc, instrument=self.instrument.upper(),
                               extrapolate_SED=self.extrapolate_SED, SED_file=self.final_SED_file,
                               SBE_save=self.source_stamps_file)
        disp_seed.observation(orders=orders)

        # Cache the dispersion if requested. This will allow you to
        #disperse the same object again later with a different lightcurve
        if cache:
            for order in orders:
                disp_seed.this_one[order].disperse_all(cache=True)
        else:
            disp_seed.disperse(orders=orders)

        # Only finalize and/or add the background if requested.
        if finalize:
            if add_background:
                background_image = disp_seed.disperse_background_1D([background_waves, background_fluxes])
                disp_seed.finalize(Back=background_image, BackLevel=None)
            else:
                disp_seed.finalize(Back=None, BackLevel=None)
        return disp_seed

    def save_seed(self, seed, segmentation_map, seed_header, params):
        """Save the 4D dispersed seed image to a fits file

        Parameters
        ----------
        seed : numpy.ndarray
            4D seed image

        segmentation_map : numpy.ndarray
            2D segmentation image

        seed_header : astropy.fits.PrimaryHDU
            Header object to use as the basis of the saved seed image

        params : dict
            Nested dictionary of instrument/observation parameters. From
            reading in a Mirage input yaml file.
        """
        arrayshape = seed.shape
        if len(arrayshape) == 4:
            units = 'ADU'
            integ, grps, yd, xd = arrayshape
            tgroup = self.frametime * (params['Readout']['nframe'] + params['Readout']['nskip'])
            print('Seed image is 4D.')
        else:
            raise ValueError('Only 4D seed images supported. This seed image is {}D'.format(len(arrayshape)))

        xcent_fov = xd / 2
        ycent_fov = yd / 2

        seed_header['XCENTER'] = xcent_fov
        seed_header['YCENTER'] = ycent_fov
        seed_header['UNITS'] = units
        seed_header['TGROUP'] = tgroup

        if params['Readout']['pupil'][0].upper() == 'F':
            usefilt = 'pupil'
        else:
            usefilt = 'filter'

        if self.total_seed_segments_and_parts == 1:
            self.seed_file = os.path.join(self.basename + '_' + params['Readout'][usefilt] +
                                          '_seed_image.fits')
        else:
            seg_string = str(self.segment_number).zfill(3)
            part_string = str(self.segment_part_number).zfill(3)
            self.seed_file = os.path.join(self.basename + '_' + params['Readout'][usefilt] +
                                          '_seg{}_part{}_seed_image.fits'.format(seg_string, part_string))

        self.seed_files.append(self.seed_file)

        # These cannot come from the values in the direct seed image, since
        # they are for the padded array.
        seed_header['NOMXDIM'] = seed.shape[1]
        seed_header['NOMYDIM'] = seed.shape[0]
        seed_header['NOMXSTRT'] = 1
        seed_header['NOMXEND'] = seed.shape[1]
        seed_header['NOMYSTRT'] = 1
        seed_header['NOMYEND'] = seed.shape[0]

        # Observations with high data volumes (e.g. moving targets, TSO)
        # can be split into multiple "segments" in order to cap the
        # maximum file size
        seed_header['EXSEGTOT'] = self.total_seed_segments  # Total number of segments in exp
        seed_header['EXSEGNUM'] = self.segment_number       # Segment number of this file
        seed_header['PART_NUM'] = self.segment_part_number  # Segment part number of this file

        # Total number of integrations and groups in the current segment
        # (combining all parts of the segment)
        seed_header['SEGINT'] = self.segment_ints
        seed_header['SEGGROUP'] = self.segment_frames

        # Total number of integrations and groups in the exposure
        seed_header['EXPINT'] = params['Readout']['nint']
        seed_header['EXPGRP'] = params['Readout']['ngroup']

        # Indexes of the ints and groups where the data in this file go
        # Frame and integration indexes of the segment within the exposure
        seed_header['SEGFRMST'] = self.segment_frame_start_number
        seed_header['SEGFRMED'] = self.segment_frame_start_number + grps - 1
        seed_header['SEGINTST'] = self.segment_int_start_number
        seed_header['SEGINTED'] = self.segment_int_start_number + integ - 1

        # Frame and integration indexes of the part within the segment
        seed_header['PTINTSRT'] = self.part_int_start_number
        seed_header['PTFRMSRT'] = self.part_frame_start_number

        # The 1b pipeline adds two header keywords to the data indicating the position
        # of the source on the detector. Currently the pipeline assumes that the
        # source is at the reference location of the aperture being used. Since Mirage
        # outputs are equivalent to what is produced by the 1b pipeline, we need to
        # add these keywords manually.
        siaf = pysiaf.Siaf('nircam')[params['Readout']['array_name']]
        seed_header['XREF_SCI'] = siaf.XSciRef
        seed_header['YREF_SCI'] = siaf.YSciRef
        seed_header['TSOVISIT'] = True

        # Save
        h0 = fits.PrimaryHDU()
        h1 = fits.ImageHDU(seed, name='DATA')
        h2 = fits.ImageHDU(segmentation_map)
        h2.header['EXTNAME'] = 'SEGMAP'

        # Put the keywords into the 0th and 1st extension headers
        for key in seed_header:
            h0.header[key] = seed_header[key]
            h1.header[key] = seed_header[key]

        hdulist = fits.HDUList([h0, h1, h2])
        hdulist.writeto(self.seed_file, overwrite=True)

        print("Seed image and segmentation map saved as {}".format(self.seed_file))
        print("Seed image, segmentation map, and metadata available as:")
        print("self.seedimage, self.seed_segmap, self.seedinfo.\n\n")

    def split_param_file(self, params):
        """Create 2 copies of the input parameter file. One will contain
        all but the TSO source, while the other will contain only the TSO
        source.

        Parameters
        ----------
        params : dict
            Nested dictionary of instrument/observation parameters. From
            reading in a Mirage input yaml file.
        """
        file_dir, filename = os.path.split(self.paramfile)
        suffix = filename.split('.')[-1]

        # Copy #1 - contains all source catalogs other than TSO sources
        # and be set to wfss mode.
        background_params = copy.deepcopy(params)
        background_params['simSignals']['tso_grism_catalog'] = 'None'
        background_params['Inst']['mode'] = 'wfss'
        background_params['Output']['grism_source_image'] = True
        background_params['Output']['file'] = params['Output']['file'].replace('.fits',
                                                                               '_background_sources.fits')
        self.background_paramfile = self.paramfile.replace('.{}'.format(suffix),
                                                           '_background_sources.{}'.format(suffix))
        utils.write_yaml(background_params, self.background_paramfile)

        # Copy #2 - contaings only the TSO grism source catalog,
        # is set to wfss mode, and has no background
        tso_params = copy.deepcopy(params)
        tso_params['Inst']['mode'] = 'wfss'
        tso_params['Output']['grism_source_image'] = True
        tso_params['simSignals']['bkgdrate'] = 0.
        tso_params['simSignals']['zodiacal'] = 'None'
        tso_params['simSignals']['scattered'] = 'None'
        other_catalogs = ['pointsource', 'galaxyListFile', 'extended', 'tso_imaging_catalog',
                          'movingTargetList', 'movingTargetSersic', 'movingTargetExtended',
                          'movingTargetToTrack']
        for catalog in other_catalogs:
            tso_params['simSignals'][catalog] = 'None'
        tso_params['simSignals']['pointsource'] = tso_params['simSignals']['tso_grism_catalog']
        tso_params['Output']['file'] = tso_params['Output']['file'].replace('.fits', '_tso_grism_sources.fits')

        self.tso_paramfile = self.paramfile.replace('.{}'.format(suffix),
                                                    '_tso_grism_sources.{}'.format(suffix))
        utils.write_yaml(tso_params, self.tso_paramfile)

    @staticmethod
    def tso_catalog_check(catalog, exp_time):
        """Check that the start and end times specified in the TSO catalog file (which are
        used to calculate the lightcurves) are long enough to encompass the entire exposure.
        If not, extend the end time to the required time.

        Parameters
        ----------
        catalog : astropy.table.Table
            Table contianing catalog of TSO source information

        exp_time : float
            Total time of the exposure, in seconds

        Returns
        -------
        catalog : astropy.table.Table
            Potentially modified catalog where the length of the lightcurve
            has been expanded to fill the entire ``exp_time``
        """
        time_unit_str = catalog['Time_units'][0]

        # Catch common unit errors
        if time_unit_str.lower() in ['seconds', 'minutes', 'hours', 'days']:
            time_unit_str = time_unit_str[0:-1]
            catalog['Time_units'][0] = time_unit_str
        catalog_time_units = u.Unit(time_unit_str)
        catalog_total_time = (catalog['End_time'][0] - catalog['Start_time'][0]) * catalog_time_units

        # Make sure the lightcurve time is at least as long as the exposure
        # time
        if exp_time > catalog_total_time:
            print(('WARNING: Lightcurve duration specified in TSO catalog file is less than '
                   'the total duration of the exposure. Adding extra time to the end of the '
                   'lightcurve to match.'))
            catalog['End_time'][0] = catalog['Start_time'][0] + catalog_total_time.value

        # Make sure the time of inferior conjunction is betwen
        # the starting and ending times
        if ((catalog['Time_of_inferior_conjunction'][0] < catalog['Start_time'][0]) or
           (catalog['Time_of_inferior_conjunction'][0] > catalog['End_time'][0])):
            raise ValueError(("ERROR: the inferior conjucntion time in the TSO catalog is "
                              "outside the bounds of the starting and ending times."))
        return catalog



"""Test the functions provided by mirage.psf.segment_psfs

Authors
-------
    - Lauren Chambers

Use
---
    >>> pytest test_segment_psfs.py
"""
from ast import literal_eval
import glob
import os
import shutil

from astropy.io import fits
import numpy as np
import photutils
import pytest
from webbpsf.utils import to_griddedpsfmodel

from .utils import parametrized_data
from mirage.psf.deployments import generate_random_ote_deployment
from mirage.psf.psf_selection import get_library_file
from mirage.psf.segment_psfs import (get_segment_library_list, get_segment_offset,
                                     get_gridded_segment_psf_library_list, generate_segment_psfs)
from mirage.utils.utils import ensure_dir_exists

# Define directory and file locations
__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
TEMP_TEST_DIRECTORY = os.path.join(__location__, 'temp_data', 'test_segment_psfs')
TEST_DATA_DIR = os.path.expandvars("$MIRAGE_DATA/test_data/test_segment_psfs")

# Load parametrized data
PARAMETRIZED_DATA = parametrized_data()['test_segment_psfs']
SUCCESS_GENERATE_SEGMENT_PSF = None

# Determine if tests are being run on Travis
ON_TRAVIS = 'travis' in os.path.expanduser('~')

# Define default inputs
INSTRUMENT = 'NIRCam'
DETECTOR = 'NRCA3'
FILTER = 'F212N'


@pytest.fixture(scope="module")
def test_directory(test_dir=TEMP_TEST_DIRECTORY):
    """Create a test directory.

    Parameters
    ----------
    test_dir : str
        Path to directory used for testing

    Yields
    -------
    test_dir : str
        Path to directory used for testing
    """
    # Create directory and yield its name
    ensure_dir_exists(test_dir)  # creates directory with default mode=511
    yield test_dir

    # Remove directory
    if os.path.isdir(test_dir):
        shutil.rmtree(test_dir)


@pytest.fixture(scope="module")
def test_library_file(test_directory):
    """Download and save locally a segment PSF library file from Box for testing

    Yields
    -------
    test_lib_filename : str
        Path to test library used for testing
    """
    # Download file and yield its name
    test_lib_filename = os.path.join(test_directory, 'test_library',
                                     'nircam_nrca3_f212n_fovp1024_samp1_npsf1_seg12.fits')
    ensure_dir_exists(os.path.dirname(test_lib_filename))
    url = "https://stsci.box.com/shared/static/4c0em1yhb1qsvrpku7j0jw1tztucvkih.fits"
    with fits.open(url) as hdulist:
        hdulist.writeto(test_lib_filename)
        yield test_lib_filename


def test_generate_segment_psfs(test_directory):
    """Test the generation of 18 segment PSFs with a randomly-perturbed
    mirror state.

    Parameters
    ----------
    test_directory : str
        Path to directory used for testing
    """
    test_data = PARAMETRIZED_DATA['test_generate_segment_psfs']

    # Randomly generate OTE state with deployment errors
    ote, segment_tilts, ote_opd_with_tilts = generate_random_ote_deployment(
        test_directory, reduction_factor=0.2
    )
    # Ensure the OPD yaml and fits files were saved
    yaml_search_path = os.path.join(test_directory, 'deployment_errors*.yaml')
    opd_search_path = os.path.join(test_directory, 'OPD*.fits')
    assert len(glob.glob(yaml_search_path)) >= 2, 'Failed to generate randomly deployed OTE'
    assert len(glob.glob(opd_search_path)) >= 2, 'Failed to generate randomly deployed OTE'

    # Ensure it doesn't work with mismatched filter & detector
    with pytest.raises(ValueError) as e:
        generate_segment_psfs(ote, segment_tilts, test_directory, filters=['F212N'],
                              detectors='NRCA5', fov_pixels=10, overwrite=False)
        assert "No matching filters and detectors given" in e, \
            'Failed to catch mismatched filter and detector'

    # Make 18 segment PSF library files
    generate_segment_psfs(ote, segment_tilts, test_directory, filters=['F212N'],
                          detectors='NRCA3', fov_pixels=101, overwrite=False)

    # Ensure that all 18 segment PSF library files were created
    fits_search_path = os.path.join(test_directory, 'nircam*.fits')
    library_names = test_data

    global SUCCESS_GENERATE_SEGMENT_PSF
    SUCCESS_GENERATE_SEGMENT_PSF = True

    for name in library_names:
        lib_success = os.path.join(test_directory, name) in glob.glob(fits_search_path)
        SUCCESS_GENERATE_SEGMENT_PSF = SUCCESS_GENERATE_SEGMENT_PSF & lib_success
        assert lib_success, 'Failed to create file: {}'.format(os.path.join(test_directory, name))


@pytest.mark.skipif(ON_TRAVIS,
                   reason="Cannot access mirage data in the central storage directory from Travis CI.")
def test_get_segment_library_list_remote():
    """Test construction of segment PSF libraries using data on the central
    storage directory.
    """
    # Get a list of segment PSF library files
    library_list = get_segment_library_list(INSTRUMENT, DETECTOR, FILTER, TEST_DATA_DIR)

    # Ensure the right files were found
    library_names = PARAMETRIZED_DATA['test_get_segment_library_list']
    assert len(library_list) == 18, 'Did not find all 18 segment libraries'
    assert sorted(library_list) == sorted([os.path.join(TEST_DATA_DIR, name) for name in library_names]), \
        'Did not find all expected 18 segment libraries'


def test_get_segment_library_list_local(test_directory):
    """Test construction of segment PSF libraries using data generated by
    test_generate_segment_psfs().

    Parameters
    ----------
    test_directory : str
        Path to directory used for testing
    """
    # Skip if test_generate_segment_psfs failed
    if not SUCCESS_GENERATE_SEGMENT_PSF:
        pytest.skip("Cannot load library files generated in test_generate_segment_psfs.")

    # Get a list of segment PSF library files
    library_list = get_segment_library_list(INSTRUMENT, DETECTOR, FILTER, test_directory)

    # Ensure the right files were found
    library_names = PARAMETRIZED_DATA['test_generate_segment_psfs']
    assert len(library_list) == 18, 'Did not find all 18 segment libraries'
    assert sorted(library_list) == sorted([os.path.join(test_directory, name) for name in library_names]), \
        'Did not find all expected 18 segment libraries'


test_data = PARAMETRIZED_DATA['test_get_segment_offset_remote']
get_segment_offset_remote_parameters = []
for i, tuple_string in enumerate(test_data):
    offset_tuple =   literal_eval(tuple_string)
    get_segment_offset_remote_parameters.append([i+1, offset_tuple])
@pytest.mark.parametrize('segment_number, correct_offset', get_segment_offset_remote_parameters)
@pytest.mark.skipif(ON_TRAVIS,
                   reason="Cannot access mirage data in the central storage directory from Travis CI.")
def test_get_segment_offset_remote(segment_number, correct_offset):
    """Test the extraction of segment offsets from segment PSF libraries on the
    central storage directory.

    Parameters
    ----------
    segment_number : int
        Segment ID, i.e 3
    correct_offset : tuple
        Expected x and y offset in arcseconds
    """
    library_path = TEST_DATA_DIR
    library_list = get_segment_library_list(INSTRUMENT, DETECTOR, FILTER, library_path)

    x_arcsec, y_arcsec = get_segment_offset(segment_number, DETECTOR, library_list)
    assert (x_arcsec, y_arcsec) == correct_offset, 'Incorrect conversion of segment offsets'


def test_get_segment_offset_local_created(test_directory):
    """Test the extraction of segment offsets from segment PSF libraries
    generated by test_generate_segment_psfs().

    Parameters
    ----------
    test_directory : str
        Path to directory used for testing
    """
    # Skip if test_generate_segment_psfs failed
    if not SUCCESS_GENERATE_SEGMENT_PSF:
        pytest.skip("Cannot load library files generated in test_generate_segment_psfs.")
    library_path = test_directory
    library_list = get_segment_library_list(INSTRUMENT, DETECTOR, FILTER, library_path)

    # Just try to get one offset and make sure it doesn't croak
    x_arcsec, y_arcsec = get_segment_offset(1, DETECTOR, library_list)
    assert type(x_arcsec) == np.float64, 'Incorrect conversion of segment offsets'
    assert type(y_arcsec) == np.float64, 'Incorrect conversion of segment offsets'


def test_get_segment_offset_local_stored(test_library_file):
    """Test the extraction of segment offsets from a single segment PSF library
    in the temp_data directory.
    """
    # Get test library file
    test_library_dir = os.path.dirname(test_library_file)
    seg_id = 12
    segment_file = get_library_file(
        INSTRUMENT, DETECTOR, FILTER, 'CLEAR', '', 0, test_library_dir,
        segment_id=seg_id
    )

    # Set segment = 0 and expect error:
    with pytest.raises(AssertionError) as e:
        library_list = [segment_file]
        get_segment_offset(0, DETECTOR, library_list)
        assert "Uh-oh. The segment ID of the library does not match the requested segment." in e, \
            'Failed to catch mismatch between segment IDs.'

    # Put segment 12th on a faked library list and extract the segment offset
    library_list = [''] * 18
    library_list[11] = segment_file
    x_arcsec, y_arcsec = get_segment_offset(12, DETECTOR, library_list)
    assert (x_arcsec, y_arcsec) == (7.532331578313969, 8.077137976712994), \
        'Incorrect conversion of segment offsets'


@pytest.mark.skipif(ON_TRAVIS,
                   reason="Cannot access mirage data in the central storage directory from Travis CI.")
def test_get_gridded_segment_psf_library_list_remote():
    """Test the loading of segment PSF libraries as a GriddedPSFModel on the
    central storage directory.
    """
    library_path = TEST_DATA_DIR

    libraries = get_gridded_segment_psf_library_list(INSTRUMENT, DETECTOR, FILTER,
                                         library_path)
    assert len(libraries) == 18, 'Did not find all 18 segment libraries'
    for i, lib_model in enumerate(libraries):
        assert isinstance(lib_model, photutils.psf.models.GriddedPSFModel), \
            'Segment PSF library not created correctly'
        assert lib_model.grid_xypos == [(1023.5, 1023.5)], \
            'Segment PSF library not created correctly'
        assert lib_model.oversampling == 1, \
            'Segment PSF library not created correctly'
        for k in ['segid', 'segname', 'xtilt', 'ytilt']:
            assert k in list(lib_model.meta.keys()), \
                'Segment PSF library not created correctly'
        assert lib_model.meta['segid'][0] == i + 1, \
            'Segment PSF library not created correctly'
        assert lib_model.data.shape == (1, 1024, 1024), \
            'Segment PSF library not created correctly'


def test_get_gridded_segment_psf_library_list_local():
    """Test the loading of a single segment PSF library as a GriddedPSFModel in
    the test_data directory.
    """
    # Skip if test_generate_segment_psfs failed
    if not SUCCESS_GENERATE_SEGMENT_PSF:
        pytest.skip("Cannot load library files generated in test_generate_segment_psfs.")

    library_path = TEMP_TEST_DIRECTORY

    libraries = get_gridded_segment_psf_library_list(INSTRUMENT, DETECTOR, FILTER,
                                         library_path)
    assert len(libraries) == 18, 'Did not find all 18 segment libraries'
    for i, lib_model in enumerate(libraries):
        assert isinstance(lib_model, photutils.psf.models.GriddedPSFModel), \
            'Segment PSF library not created correctly'
        assert lib_model.grid_xypos == [(1023.0, 1023.0)], \
            'Segment PSF library not created correctly'
        assert lib_model.oversampling == 1, \
            'Segment PSF library not created correctly'
        for k in ['segid', 'segname', 'xtilt', 'ytilt']:
            assert k in list(lib_model.meta.keys()), \
                'Segment PSF library not created correctly'
        assert lib_model.meta['segid'][0] == i + 1, \
            'Segment PSF library not created correctly'
        assert lib_model.data.shape == (1, 101, 101), \
            'Segment PSF library not created correctly'


def test_to_gridded_psfmodel(test_library_file):
    """Test that the example library file can be correctly loaded as a
    GriddedPSFModel using the webbpsf.utils.to_griddedpsfmodel function.

    Note that this is more a test of webbpsf than of MIRaGe, but it is
    required for MIRaGe to work!
    """
    with fits.open(test_library_file) as hdulist:
        lib_model = to_griddedpsfmodel(hdulist)

    assert isinstance(lib_model, photutils.psf.models.GriddedPSFModel), \
        'Segment PSF library not created correctly'
    assert lib_model.grid_xypos == [(1023.5, 1023.5)], \
        'Segment PSF library not created correctly'
    assert lib_model.oversampling == 1, \
        'Segment PSF library not created correctly'
    for k in ['segid', 'segname', 'xtilt', 'ytilt']:
        assert k in list(lib_model.meta.keys()), \
            'Segment PSF library not created correctly'
    assert lib_model.meta['segid'][0] == 12, \
        'Segment PSF library not created correctly'
    assert lib_model.data.shape == (1, 1024, 1024), \
        'Segment PSF library not created correctly'



#! /usr/bin/env python

"""
Create a mirage-format seed image from
an input fits file. This fits file must contain a
valid WCS, and be distortion-free.

This code will extract a sub-image from the input
fits file. The subimage is centered at the input
RA and Dec (crop_center_ra, crop_center_dec) and
its size is determined by the aperture input.

The extracted sub-image is then blotted (resampled)
onto the requested NIRCam detector's FOV, including
the distortion model.

The resulting image is then saved in such a format
that it can be used as input in the obs_generator.py,
image_simulator.py, or disperser code in the
mirage package. Essentially this script
is designed to take the place of catalog_seed_image.py
in the case where the user has a distortion-free
mosaic image that they wish to break up into
NIRCam images.

Another way to use this code would be for an input
fits image of some extended object that the user
wishes to distort according to the NIRCam distortion
model. This distorted file can then be placed in
an extended object catalog which will be used by
the catalog_seed_image.py step. In this method,
multiple extended objets could be distorted and
added to the same frame.

Inputs:
mosaic_file: The name of the fits file containing the
            distortion-free image to be blotted. The code
            currently assumes that the mosaic_file is
            oriented north-up.

From here, there are two ways to call fits_seed_image.py

1. parameter file: a yaml input file matching the format
   (optional)      required by other steps of the nircam
                   data simulator

2. Manual inputs:
   aperture -
              aperture name matching an entry in the
              subarray definition file (e.g. NRCB5_FULL)

   crop_center_ra, crop_center_dec -
              The RA and Dec at the center of the sub-image
              to crop from mosaic_file. This, in combination
              with the array size (determined from aperture)
              define the sub-image to use.

   blot_center_ra, blot_center_dec -
              The RA and Dec at the center of the blotted
              sub-image. If these are equal to crop_center_ra
              and dec, then the center of the cropped image
              will remain at the center of the blotted image.

   blot_pav3 -
              Position angle of the blotted image.

   flux_cal_file -
              Ascii file listing the zeropoints for all
              NIRCam filters. By setting this value equal
              to 'config', the code will use the flux
              calibration file that is packaged with the
              NIRCam data simulator. This information is
              used only to populate the PHOTFLAM, PHOTFNU,
              and PHOTPLAM header keywords.

   filter -
              The NIRCam filter to use. This is used to
              select the appropriate row from the flux
              calibration file.

   pupil -
              The NIRCam pupil value to use. This is used
              to select the appropriate row from the flux
              calibration file in the case where a filter
              that is located in the pupil wheel is used.
              If the pupil value does not start with 'F'
              (e.g. F405N), then it is ignored.

   grism_source_image -
              True/False. If you intend to send the seed
              image output to the disperser software, then
              set this value to True. The output image will
              then be made larger than the nominal aperture
              size by a factor of sqrt(2), in order for the
              disperser to keep track of sources just outside
              the nominal FOV, whose dispersed spectra may
              fall onto the detector.

   outfile -
              Name of the file to contain the output seed
              image.

   outdir -
              Directory in which to place outfile

Outputs:
self.seedimage, self.seed_segmap, self.seedinfo
contain the seed (countrate) image, the associated
segmentation map, and header information required
by subsequent steps of the nircam data simulator.
The segmentation map is only used in the case where
the seed image is dispersed by a call to the
disperser software.

Example calls:
See the Simulated_data_from_mosaic_image.ipynb notebook
in the examples directory
"""
import argparse
import copy
import os
import pkg_resources
import sys

from astropy.io import fits, ascii
import datetime
from math import radians
import numpy as np
from photutils import detect_sources
from photutils import TopHatWindow, TukeyWindow, CosineBellWindow, SplitCosineBellWindow, HanningWindow
from photutils.centroids import centroid_2dg
from photutils.psf import resize_psf
from photutils.psf.matching import create_matching_kernel
import pysiaf
from scipy.ndimage import shift
from scipy.signal import fftconvolve
import yaml

from . import crop_mosaic, blot_image
from mirage.psf.psf_selection import get_psf_wings
from mirage.psf import tools
from mirage.seed_image.save_seed import save
from mirage.reference_files import crds_tools
from mirage.utils.constants import EXPTYPES
from mirage.utils.siaf_interface import get_siaf_information


config_files = {'nircam': {'flux_cal': 'NIRCam_zeropoints.list'},
                'niriss': {'flux_cal': 'niriss_zeropoints.list'}
                }

KNOWN_PSF_TELESCOPES = {"JWST", "HST", "SPITZER"}


class ImgSeed:
    def __init__(self, paramfile=None, mosaic_file=None, data_extension_number=0, wcs_extension_number=0,
                 cropped_file='cropped_image.fits', outdir=None, blotted_file=None, psf_file=None,
                 mosaic_fwhm=None, mosaic_fwhm_units='arcsec', gaussian_psf=True, save_intermediates=False):
        """Create a seed image from a distortionless input mosaic. The user can supply
        a PSF associated with the mosaic in the ``psf_file`` parameter. If ``psf_file``
        is None and ``gaussian_psf`` is True, then astropy's Gaussian2D model will be
        used to construct a PSF for the mosaic. If ``psf_file`1` is None and ``gaussian_psf``
        is False, and the mosaic comes from a "known" telescope, then a custom function
        will be used to construct the PSF. Currently the "known" telescopes include
        JWST, HST, and Spitzer IRAC.

        Note that this code cannot be used in cases where the mosaic's PSF FWHM is
        larger than the JWST PSF, as translating the mosaic to the JWST PSF would then
        be sharpening the image.

        Parameters
        ----------
            paramfile : str
                Name of Mirage yaml input file used to supply instrument info

            mosaic_file : str
                Name of the fits file containing the mosaic file to use

            data_extension_number : int
                Extension number within the mosaic FITS file where the data
                array is located

            wcs_extension_number : int
                Extension number within the mosaic FITS file where the
                world coordinate system information is located

            cropped_file : str
                Name of the file to save the cropped image into. If None,
                the cropped image is not saved

            outdir : str
                Name of the output directory to save the output files into.
                If None, the output directory listed in ``paramfile`` will
                be used.

            blotted_file : str
                Name of FITS file to save resampled image into

            psf_file : str
                Name of FITS file containing a PSF corresponding to the mosaic
                data. This will be used to create a matching kernel to allow
                convolution of the mosaic data to match the JWST PSF. If None,
                a PSF will be created using an astropy Gaussian2D model or,
                in the case of a mosaic from JWST, a webbpsf-generated PSF
                will be read in.

            mosaic_fwhm : float
                FWHM of the PSF in the mosaic. Can be in pixels or arcseconds,
                as specified by the ``mosaic_fwhm_units`` keyword. If None,
                an astropy Gaussian2D model will be fit to the PSF to estimate
                the FWHM

            mosaic_fwhm_units : str
                Units of ``mosaic_fwhm``. Can be 'pixels' or 'arcsec'

            gaussian_psf : bool
                If ``psf_file`` is None and this is True, the mosaic PSF will
                be created using an astropy Gaussian2D model.

            save_intermediates : bool
                If True, intermediate outputs, including the PSFs and the
                matching PSF kernel, as well as the cropped image after
                convolution with the matching PSF kernel, will be saved.
        """
        allowed_mosaic_fwhm_units = ['pixels', 'arcsec']

        self.mosaic_file = mosaic_file
        self.data_extension_number = data_extension_number
        self.wcs_extension_number = wcs_extension_number
        self.crop_center_ra = 0.
        self.crop_center_dec = 0.
        self.channel = None
        self.detector = None
        self.blot_center_ra = 0.
        self.blot_center_dec = 0.
        self.blot_pav3 = 0.
        self.aperture = ''
        self.flux_cal_file = 'config'
        self.distortion_file = 'crds'
        self.filter = ''
        self.pupil = ''
        self.grism_source_image = False
        self.cropped_file = cropped_file
        self.blotted_file = blotted_file
        self.outdir = outdir
        self.grism_direct_factor = np.sqrt(2.)
        self.psf_file = psf_file
        self.gaussian_psf = gaussian_psf
        self.mosaic_fwhm = mosaic_fwhm
        if mosaic_fwhm_units not in allowed_mosaic_fwhm_units:
            raise ValueError(("ERROR: mosaic_fwhm_units must be one of: {}"
                              .format(allowed_mosaic_fwhm_units)))
        self.mosaic_fwhm_units = mosaic_fwhm_units
        self.save_intermediates = save_intermediates

        # Locate the module files, so that we know where to look
        # for config subdirectory
        self.modpath = pkg_resources.resource_filename('mirage', '')

        # self.coords contains the factor by which the
        # nominal output array size needs to be increased
        # (used for WFSS mode), as well as the coordinate
        # offset between the nominal output array coordinates,
        # and those of the expanded array. These are needed
        # mostly for WFSS observations, where the nominal output
        # array will not sit centered in the expanded output image.
        self.coords = {'x': 1., 'xoffset': 0., 'y': 1., 'yoffset': 0.}

        # If a paramfile is provided, read in and
        # set params
        self.paramfile = paramfile
        if paramfile is not None:
            self.read_param_file(paramfile)

    def aperture_info(self):
        """
        Get basic information on the reuqested aperture from SIAF
        """
        local_roll, attitude_matrix, self.framesize, subarray_bounds = get_siaf_information(self.siaf,
                                                                                            self.aperture,
                                                                                            self.crop_center_ra,
                                                                                            self.crop_center_dec,
                                                                                            self.blot_pav3)
        xstart, ystart, xend, yend = subarray_bounds

        self.subarr_bounds = {'xstart': xstart, 'ystart': ystart, 'xend': xend, 'yend': yend}

        self.nominal_dims = np.array([yend - ystart + 1, xend - xstart + 1])

    def crop_and_blot(self):
        """MAIN FUNCTION

        Extract an area of the input mosaic fits file that is slightly larger
        than the desired aperture, and use the JWST pipeline version of blot (called
        resample) to resample this to the correct pixel scale and introduce the
        proper distortion for the requested detector
        """
        self.detector_channel_value()
        self.distortion_file_value()
        module = self.module_value()
        self.siaf = pysiaf.Siaf(self.instrument)

        # Get information on the aperture
        self.aperture_info()
        xstart = self.subarr_bounds['xstart']
        ystart = self.subarr_bounds['ystart']
        xdim = self.subarr_bounds['xend'] - xstart + 1
        ydim = self.subarr_bounds['yend'] - ystart + 1

        # Flux calibration information
        self.flux_cal_file = self.expand_config_paths(self.flux_cal_file, 'flux_cal')
        if self.pupil[0].upper() == 'F':
            usefilt = self.pupil.upper()
        else:
            usefilt = self.filter.upper()
        self.fluxcal_info(usefilt, module)

        # Get information on the instrument used to create the mosaic
        self.mosaic_metadata = tools.get_psf_metadata(self.mosaic_file)

        # Check that the FWHM of the mosaic PSF is smaller than the PSF
        # of the JWST PSF. If the opposite is true, we can't create a
        # matching kernel
        if self.mosaic_fwhm_units == 'arcsec':
            mosaic_fwhm_arcsec = copy.deepcopy(self.mosaic_fwhm)
            self.mosaic_fwhm /= self.mosaic_metadata['pix_scale1']

        # Identify the correct PSF for the JWST instrument/detector
        # Let's just read in the psf_wings file, which contains the PSF
        # at the center of the detector. It doesn't make much sense to
        # worry about spatially-dependent PSFs in this case
        self.jwst_psf = get_psf_wings(self.instrument, self.detector, self.filter, self.pupil,
                                      self.params['simSignals']['psfwfe'],
                                      self.params['simSignals']['psfwfegroup'],
                                      os.path.join(self.params['simSignals']['psfpath'], 'psf_wings'))
        self.outscale1, self.outscale2 = self.find_jwst_pixel_scale()

        # Find the FWHM of the JWST PSF
        j_yd, j_xd = self.jwst_psf.shape
        mid_y = j_yd // 2
        mid_x = j_xd // 2
        box = self.jwst_psf[mid_y-10:mid_y+11, mid_x-10:mid_x+11]
        jwst_x_fwhm, jwst_y_fwhm = tools.measure_fwhm(box)
        jwst_y_fwhm_arcsec = jwst_y_fwhm * self.outscale2
        print('JWST FWHM in pix: ', jwst_y_fwhm)
        print('JWST FWHM in arcsec: ', jwst_y_fwhm_arcsec)

        # If the FWHM of the mosaic image is larger than that of the JWST
        # PSF, then we cannot continue because we cannot create a matching
        # PSF kernel to use for convolution
        if mosaic_fwhm_arcsec > jwst_y_fwhm_arcsec:
            raise ValueError(("ERROR: FWHM of the mosaic image is larger than that of "
                              "the JWST PSF. Unable to create a matching PSF kernel "
                              "using photutils. \nMosaic FWHM: {}\nJWST FWHM: {}"
                              .format(self.mosaic_fwhm, jwst_y_fwhm_arcsec)))

        # The JWST PSF as read in is large (399 x 399 pixels). Crop to
        # something reasonable so that the convolution doesn't take
        # forever
        half_width = 50
        self.jwst_psf = self.jwst_psf[mid_y-half_width: mid_y+half_width+1, mid_x-half_width: mid_x+half_width+1]

        # Collect the metadata relating to the mosaic and (optionally) PSF
        # and read in/create the PSF
        self.prepare_psf()

        # Crop
        print("Cropping subarray from mosaic")
        pixel_scale = self.siaf[self.aperture].XSciScale
        crop = crop_mosaic.Extraction(mosaicfile=self.mosaic_file,
                                      data_extension_number=self.data_extension_number,
                                      wcs_extension_number=self.wcs_extension_number,
                                      center_ra=self.crop_center_ra,
                                      center_dec=self.crop_center_dec,
                                      dimensions=(xdim, ydim),
                                      outfile=self.cropped_file,
                                      jwst_pixel_scale=pixel_scale)
        crop.extract()

        # Convolve the cropped image with the appropriate PSF
        print("Convolving with PSF kernel")
        crop.cropped = self.psf_convolution(crop.cropped)

        # Blot
        print("Resampling subarray onto JWST pixel grid")
        blot = blot_image.Blot(instrument=self.instrument, aperture=self.aperture,
                               ra=[self.blot_center_ra], dec=[self.blot_center_dec],
                               pav3=[self.blot_pav3], blotfile=crop.cropped,
                               distortion_file=self.distortion_file)
        blot.blot()

        # Blot the PSF associated with the mosaic
        #blot_psf = blot_image.Blot(instrument=self.instrument, aperture=self.aperture,
        #                           ra=[self.blot_center_ra], dec=[self.blot_center_dec],
        #                           pav3=[self.blot_pav3], blotfile=mosaic_psf_model,
        #                           distortion_file=self.distortion_file)
        #blot_psf.blot()

        for dm in blot.blotted_datamodels:
            # Create segmentation map
            # (used only when dispsersing image for WFSS)
            self.seed_image = dm.data
            self.seed_segmap = self.make_segmap(dm)

            # Save
            if self.blotted_file is not None:
                self.blotted_file = os.path.join(self.outdir, self.blotted_file)
            self.seed_file, self.seedinfo = save(dm.data, self.paramfile, self.params,
                                                 self.photflam, self.photfnu, self.pivot,
                                                 self.framesize, self.nominal_dims,
                                                 self.coords, self.grism_direct_factor,
                                                 segmentation_map=self.seed_segmap,
                                                 filename=self.blotted_file, base_unit='e-')
            print('Blotted image saved to: {}'.format(self.seed_file))

    def detector_channel_value(self):
        """Get the detector and optional channel value based on the aperture
        name
        """
        if self.detector is None:
            self.detector = self.aperture[0:5]
        self.detector = self.detector.upper()
        self.instrument = self.instrument.upper()

        # Make sure the FGS detector is GUIDER[12]
        if 'FGS' in self.detector:
            self.detector = 'GUIDER{}'.format(self.detctor[-1])

        if 'NRC' in self.detector:
            if self.channel is None:
                if '5' in self.detector:
                    self.channel = 'LONG'
                else:
                    self.channel = 'SHORT'

    def distortion_file_value(self):
        """Make sure the distortion file name is a valid file. If it has a
        value of 'crds', then query CRDS and retrieve the appropriate file
        """
        if self.distortion_file == 'crds':
            if self.paramfile is None:
                self.reference_dictionary = {}
                self.reference_dictionary['INSTRUME'] = self.instrument
                self.reference_dictionary['DETECTOR'] = self.detector
                if '5' in self.detector:
                    self.reference_dictionary['DETECTOR'] = self.detector.replace('5', 'LONG')

                if self.instrument.upper() == 'NIRCAM':
                    self.reference_dictionary['CHANNEL'] = self.channel

                if 'FGS' in self.reference_dictionary['DETECTOR']:
                    self.reference_dictionary['DETECTOR'] = 'GUIDER{}'.format(self.reference_dictionary['DETECTOR'][-1])

                # Use the current date and time in order to get the most recent
                # reference file
                self.reference_dictionary['DATE-OBS'] = datetime.date.today().isoformat()
                current_date = datetime.datetime.now()
                self.reference_dictionary['TIME-OBS'] = current_date.time().isoformat()

                # For the purposes of choosing reference files, the exposure type should
                # always be set to imaging, since it is used to locate sources in the
                # seed image, prior to any dispersion.
                self.reference_dictionary['EXP_TYPE'] = EXPTYPES[self.instrument.lower()]["imaging"]

                # This assumes that filter and pupil names match up with reality,
                # as opposed to the more user-friendly scheme of allowing any
                # filter to be in the filter field.
                self.reference_dictionary['FILTER'] = self.filter
                self.reference_dictionary['PUPIL'] = self.pupil

            mapping = crds_tools.get_reffiles(self.reference_dictionary, ['distortion'], download=True)
            self.distortion_file = mapping['distortion']

    def expand_config_paths(self, filename, filetype):
        """If a reference file has the name "config" in the input yaml
        file, then convert that to the appropriate file name. This is
        done using a global dictionary.

        Parameters
        ----------
        filename : str
            The filename as listed in the yaml file

        filetype : str
            The type of reference file in question. Currently
            the dictionary only has an entry for 'flux_cal'

        Returns
        -------
        filename : str
            The updated filename for the reference file type in question
        """
        if filename.lower() == 'config':
            sfile = config_files[self.instrument.lower()][filetype]
            return os.path.join(self.modpath, 'config', sfile)
        else:
            return filename

    def find_jwst_pixel_scale(self):
        """Find the pixel scale for the JWST instrument and aperture to
        be used for the seed image
        """
        meta_output = {}
        meta_output['instrument'] = self.instrument
        return tools.get_JWST_pixel_scale(meta_output, aperture=self.aperture)

    def fluxcal_info(self, filter_name, module):
        """Get the flux calibration-related information from the flux
        cal reference file.

        Parameters
        ----------
        filter_name : str
            Name of the filter to look up the flux cal info for

        module : str
            Module name to use for the look-up 'A' or 'B' for
            NIRCam, and 'N' for NIRISS
        """
        # Read in list of zeropoints/photflam/photfnu
        self.zps = ascii.read(self.flux_cal_file)
        mtch = ((self.zps['Filter'] == filter_name) \
                & (self.zps['Module'] == module))
        self.photflam = self.zps['PHOTFLAM'][mtch][0]
        self.photfnu = self.zps['PHOTFNU'][mtch][0]
        self.pivot = self.zps['Pivot_wave'][mtch][0]

    def grism_coord_adjust(self):
        """Calculate the factors by which to expand the
        output array size, as well as the coordinate
        offsets between the nominal output array and
        the input lists if the observation being
        modeled is to be dispersed with the grism
        """
        dtor = radians(1.)

        # Normal imaging with grism image requested
        self.coords['x'] = self.grism_direct_factor
        self.coords['y'] = self.grism_direct_factor
        self.coords['xoffset'] = np.int((self.grism_direct_factor - 1.)
                                        * (self.subarr_bounds['xend'] -
                                           self.subarr_bounds['xstart'] + 1) / 2.)
        self.coords['yoffset'] = np.int((self.grism_direct_factor - 1.)
                                        * (self.subarr_bounds['yend'] -
                                           self.subarr_bounds['ystart']+1) / 2.)

    def make_segmap(self, model):
        """Create a segmentation map of the input image

        Parameters
        ----------
        model : jwst.datamodels.ImageModel

        Returns
        -------
        map_image : numpy.ndarray
            2D segmentation map
        """
        noise = np.median(model.data)
        seg_map = detect_sources(model.data, noise*5., 8)
        if seg_map is None:
            print('No segmentation map created. Returning empty segmap')
            map_image = np.zeros_like(model.data)
        else:
            map_image = seg_map.data
        return map_image

    @staticmethod
    def matching_kernel(psf1, psf2, window_type='TukeyWindow', alpha=None, beta=None):
        """Use photutils to create a matching kernel given two PSFs and
        the window type and parameters

        Parameters
        ----------
        psf1 : numpy.ndarray
            2D array containing the first PSF

        psf2 : numpy.ndarray
            2D array containing the second PSF

        window_type : str
            Name of the window function to use when filtering the matching kernel

        alpha : float
            Optional input for some of the window functions

        beta : float
            Optional input for some of the window functions

        Returns
        -------
        matched_kernel : numpy.ndarray
            2D array containing the matching PSF kernel
        """
        # Create the filtering window
        orig_window_type = copy.deepcopy(window_type)
        window_type = window_type.lower()
        if window_type == 'tophatwindow':
            window = TopHatWindow(beta=beta)
        elif window_type == 'cosinebellwindow':
            window = CosineBellWindow(alpha=alpha)
        elif window_type == 'splitcosinebellwindow':
            window = SplitCosineBellWindow(alpha=alpha, beta=beta)
        elif window_type == 'tukeywindow':
            window = TukeyWindow(alpha=alpha)
        elif window_type == 'hanningwindow':
            window = HanningWindow()
        else:
            raise ValueError("ERROR: Unrecognized window_type: {}".format(orig_window_type))

        # Create the matching kernel
        matched_kernel = create_matching_kernel(psf1, psf2, window=window)

        return matched_kernel

    def module_value(self):
        """Determine the module value based on the instrument and detector
        """
        if self.instrument == 'NIRCAM':
            module = self.detector[3].upper()
        elif self.instrument == 'NIRISS':
            module = 'N'
        elif self.instrument == 'FGS':
            module = 'G'
        return module

    def prepare_psf(self):
        """Get metadata relevant to the instrument/detector as well as the
        pixel scale from either the mosaic file, the PSF file, or both.
        Read in or create the PSF that is representative of the mosaic
        data.
        """
        # Get the PSF associated with the mosaic
        if self.psf_file is None:

            # If no PSF file is given and there is no pixel scale listed in
            # the mosaic file, then we cannot continue
            if self.mosaic_metadata['pix_scale1'] is None:
                raise ValueError(("ERROR: pixel scale value not present in mosaic file "
                                  "(in CD1_1 header keyword). This information is needed "
                                  "to be able to convolve the mosaic with the proper PSF "
                                  "kernel."))

            # Get the dimensions of the JWST PSF kernel representing the
            # final PSF to use in the simulation
            jwst_ydim, jwst_xdim = self.jwst_psf.shape

            # If no psf file is given, and the user requests a 2D Gaussian,
            # then make a 2D Gaussian
            if self.gaussian_psf:
                print("Creating 2D Gaussian for mosiac PSF")

                # If a Gaussian FWHM value is given, then construct the
                # PSF using astropy's Gaussian2D kernel

                # Rough guess on dimensions to use
                scale_ratio1 = self.outscale1 / self.mosaic_metadata['pix_scale1']
                scale_ratio2 = self.outscale2 / self.mosaic_metadata['pix_scale2']
                gauss_xdim = int(np.round(jwst_xdim * scale_ratio1))
                gauss_ydim = int(np.round(jwst_ydim * scale_ratio2))

                # Make sure the array has an odd number of rows and columns
                if gauss_xdim % 2 == 0:
                    gauss_xdim += 1
                if gauss_ydim % 2 == 0:
                    gauss_ydim += 1

                print('Mosaic pixel scale:', self.mosaic_metadata['pix_scale1'], self.mosaic_metadata['pix_scale2'])
                print('JWST pixel scale:', self.outscale1, self.outscale2)
                print('scale ratios:', scale_ratio1, scale_ratio2)
                print('JWST PSF dims:', jwst_xdim, jwst_ydim)
                print('Gaussian PSF dimensions: ', gauss_xdim, gauss_ydim)


                # Create 2D Gaussian
                self.mosaic_psf = tools.gaussian_psf(self.mosaic_fwhm, gauss_xdim, gauss_ydim)

                print('Temporarily saving psf for development')
                h0 = fits.PrimaryHDU(self.mosaic_psf)
                hlist = fits.HDUList([h0])
                hlist.writeto('gaussian_2d_psf.fits', overwrite=True)
            elif self.mosaic_metadata['telescope'] in KNOWN_PSF_TELESCOPES and not self.gaussian_psf:
                # If no PSF file is given and the user does not want a
                # generic 2D Gaussian (self.gaussian_psf is False),
                # check to see if the mosaic is from one of the telescopes
                # with a known PSF. If so, use the appropriate function to
                # construct a PSF
                if self.mosaic_metadata['telescope'] == 'HST':
                    print('Creating HST PSF, using 2D Gaussian')
                    print('HST FWHM in arcsec: {}'.format(self.mosaic_fwhm * self.mosaic_metadata['pix_scale2']))
                    self.mosaic_psf = tools.get_HST_PSF(self.mosaic_metadata, self.mosaic_fwhm)
                elif self.mosaic_psf_metadata['telescope'] == 'JWST':
                    print("Retrieving JWST PSF")
                    self.mosaic_psf = tools.get_JWST_PSF(self.mosaic_metadata)
                elif self.mosaic_psf_metadata['instrument'] == 'IRAC':
                    print("Retrieving IRAC PSF")
                    self.mosaic_psf = tools.get_IRAC_PSF(self.mosaic_metadata)
            else:
                raise ValueError(("For telescopes other than {}, you must either provide a "
                                  "representative PSF using the psf_file keyword, or set "
                                  "gaussian_psf=True in order to construct a 2D Gaussian PSF."
                                  "This is neeeded to create the proper PSF kernal to transform "
                                  "to the appropriate JWST PSF."))
        else:
            # If a PSF file is provided, check for any metadata. Metadata
            # from the mosaic takes precidence over metadata in the PSF file.
            psf_metadata = tools.get_psf_metadata(psf_filename)

            # If the mosaic has no pixel scale info but the PSF file does,
            # use the value from the PSF file.
            if self.mosaic_metadata['pix_scale1'] is None:
                if psf_metadata['pix_scale1'] is not None:
                    self.mosaic_metadata['pix_scale1'] = psf_metadata['pix_scale1']
                else:
                    raise ValueError(("ERROR: no pixel scale value present in mosaic file nor PSF "
                                      "file metadata (in CD1_1 header keyword). This information is "
                                      "needed to be able to convolve the mosaic with the proper PSF "
                                      "kernel."))
            self.mosaic_psf = fits.getdata(psf_filename)

    def psf_convolution(self, model):
        """Convolve the cropped image with the appropriate PSF for the
        JWST detector being simulated.

        Parameters
        ----------
        model : jwst.datamodels.ImageModel
            Data model instance containing the cropped image

        Returns
        -------
        model : jwst.datamodels.ImageModel
            Data model with image convolved by the PSF
        """
        # The jwst_psf and the mosaic_psf must have the same array size
        # and the same pixel scale. First deal with the pixel scale.

        # Rescale one of the PSFs if necessary, in order to get matching pixel scales.
        # Since the matching kernel is going to be convolved with the mosaic image,
        # then it seems like we should match the PSFs at the mosaic pixel scale.
        if not np.isclose(self.outscale1, self.mosaic_metadata['pix_scale1'], atol=0., rtol=0.01):
            orig_jwst = copy.deepcopy(self.jwst_psf)
            self.jwst_psf = resize_psf(self.jwst_psf, self.outscale1, self.mosaic_metadata['pix_scale1'], order=3)

            resized_y_dim, resized_x_dim = self.jwst_psf.shape
            if ((resized_y_dim % 2 == 0) or (resized_x_dim % 2 == 0)):
                if resized_y_dim % 2 == 0:
                    new_y_dim = resized_y_dim + 1
                else:
                    new_y_dim = resized_y_dim
                if resized_x_dim % 2 == 0:
                    new_x_dim = resized_x_dim + 1
                else:
                    new_x_dim = resized_x_dim

                # Add a column/row to the resized array,
                jwst_psf_padded = np.zeros((new_y_dim, new_x_dim))
                jwst_psf_padded[0: resized_y_dim, 0: resized_x_dim] = self.jwst_psf
                # Rather than zeros, make the top row/leftmost column a
                # copy of the row/column next to it
                if new_y_dim > resized_y_dim:
                    jwst_psf_padded[-1, 0: resized_x_dim] = self.jwst_psf[-1, :]
                if new_x_dim > resized_x_dim:
                    jwst_psf_padded[0: resized_y_dim, -1] = self.jwst_psf[:, -1]
                if ((new_y_dim > resized_y_dim) and (new_x_dim > resized_x_dim)):
                    jwst_psf_padded[-1, -1] = self.jwst_psf[-1, 1]

                # Shift the source to be centered in the center pixel
                centerx, centery = centroid_2dg(jwst_psf_padded)
                jwst_shifted = shift(jwst_psf_padded, [0.5, 0.5], order=1)
                centerx, centery = centroid_2dg(jwst_shifted)
                self.jwst_psf = jwst_shifted
            else:
                jwst_psf_padded = self.jwst_psf

        jwst_shape = self.jwst_psf.shape
        mosaic_shape = self.mosaic_psf.shape
        if ((jwst_shape[0] % 2 != mosaic_shape[0] % 2) or (jwst_shape[1] % 2 != mosaic_shape[1] % 2)):
            raise ValueError(("ERROR: Mosaic PSF and JWST PSF have different shapes in terms "
                              "of odd/even numbers of rows and/or columns. Try adding or subtracting "
                              "rows/columns to the mosaic PSF. Mosaic PSF shape: {}, JWST PSF shape: {}"
                              .format(mosaic_shape, jwst_shape)))

        # Now crop either the resized JWST PSF or the mosaic PSF in
        # order to get them both to the same array size
        print("Crop PSFs to have the same array size")
        self.jwst_psf, self.mosaic_psf = tools.same_array_size(self.jwst_psf, self.mosaic_psf)

        # Now we make a matching kernel. The mosaic can then be
        # convolved with this kernel in order to adjust the PSFs to match
        # those from JWST.
        print("Create matching kernel")
        kernel = self.matching_kernel(self.mosaic_psf, self.jwst_psf, window_type='TukeyWindow',
                                      alpha=1.5, beta=1.5)

        if self.save_intermediates:
            print('Save JWST psf and matching psf in outgoing_and_matching_kernel.fits')
            ha = fits.PrimaryHDU(orig_jwst)
            h0 = fits.ImageHDU(self.jwst_psf)
            h1 = fits.ImageHDU(self.mosaic_psf)
            h2 = fits.ImageHDU(kernel)
            hlist = fits.HDUList([ha, h0, h1, h2])
            outfile = os.path.join(self.outdir,
                                   '{}_outgoing_and_matching_kernel.fits'.format(self.output_base))
            hlist.writeto(outfile, overwrite=True)

        print('Convolve image cropped from mosaic with the matching PSF kernel')
        start_time = datetime.datetime.now()
        convolved_mosaic = fftconvolve(model.data, kernel, mode='same')
        end_time = datetime.datetime.now()
        delta_time = end_time - start_time
        print("Convolution took {} seconds".format(delta_time.seconds))
        model.data = convolved_mosaic

        if self.save_intermediates:
            print('Saving convolved mosaic as convolved_mosaic.fits')
            h0 = fits.PrimaryHDU(convolved_mosaic)
            hlist = fits.HDUList([h0])
            outfile = os.path.join(self.outdir,
                                   '{}_convolved_mosaic.fits'.format(self.output_base))
            hlist.writeto(outfile, overwrite=True)

        return model

    def read_param_file(self, file):
        """Read the input yaml file into a nested dictionary

        Parameters
        ----------
        file : str
            Name of the input yaml file
        """
        if os.path.isfile(file):
            with open(file, 'r') as f:
                self.params = yaml.safe_load(f)
        else:
            raise FileNotFoundError(("ERROR: {} does not exist."
                                     .format(file)))

        self.instrument = self.params['Inst']['instrument'].upper()
        self.aperture = self.params['Readout']['array_name'].upper()
        self.detector_channel_value()

        self.crop_center_ra = self.params['Telescope']['ra']
        self.crop_center_dec = self.params['Telescope']['dec']
        self.blot_center_ra = self.params['Telescope']['ra']
        self.blot_center_dec = self.params['Telescope']['dec']
        self.blot_pav3 = self.params['Telescope']['rotation']

        self.flux_cal_file = self.params['Reffiles']['flux_cal']
        self.distortion_file = self.params['Reffiles']['astrometric']
        self.filter = self.params['Readout']['filter']
        self.pupil = self.params['Readout']['pupil']

        self.output_base = self.params['Output']['file']

        # If input is to be dispersed later, we need to
        # expand the region to be blotted
        if self.params['Output']['grism_source_image']:
            self.grism_coord_adjust()

        if self.outdir is None:
            self.outdir = self.params['Output']['directory']

        # Create a dictionary for future CRDS query
        self.reference_dictionary = crds_tools.dict_from_yaml(self.params)

    def add_options(self, parser=None, usage=None):
        if parser is None:
            parser = argparse.ArgumentParser(usage=usage, \
                                             description='Create seed image via catalogs')
        parser.add_argument("paramfile", help='File describing the input parameters and instrument settings to use. (YAML format).')
        parser.add_argument("mosaic_file", help="Fits file containing the image to create the seed image from")
        parser.add_argument("data_extension_number", help="Extension in fits file that contains mosaic image", default=0)
        parser.add_argument("wcs_extension_number", help="Extension in mosaic fits file that contains WCS information", default=0)
        parser.add_argument("cropped_file", help="Filename used to save image cropped from mosaic", default=None)
        parser.add_argument("blotted_file", help="Filename used to save resampled image", default=None)
        parser.add_argument("outdir", help="Output directory. If None, the output dir from paramfile is used.", default=None)
        parser.add_argument("psf_file", help="Name of fits file containing PSF corresponding to mosaic image", default=None)
        parser.add_argument("mosaic_fwhm", help="FWHM of PSF in mosaic image, in pixels or arcsec", default=None)
        parser.add_argument("mosaic_fwhm_units", help="Units of PSF FWHM in mosaic image. 'pixels' or 'arcsec'", default='arcsec')
        parser.add_argument("gaussian_psf", help="If True, create mosaic PSF from Gaussian2D model", default=False, action='store_true')
        parser.add_argument("save_intermediates", help="If True, save intermediate outputs", default=False, action='store_true')
        return parser


if __name__ == '__main__':

    usagestring = 'USAGE: fits_seed_image.py inputs.yaml'
    seed = Imgseed()
    parser = seed.add_options(usage=usagestring)
    args = parser.parse_args(namespace=seed)
    seed.crop_and_blot()



'''Create necessary catalog files for targets in a given APT proposal.

Extracts the RA and Dec of all observed targets from the APT .pointing
file, and creates a catalog file that lists the sources within a
defined radius of each of these targets in mirage/catalogs/.
This module uses astroquery and Vizier to query 2MASS and WISE for
short-wave (F212N) and long-wave (F480M) catalogs, respectively, and
writes .list files.

Authors
-------
    - Lauren Chambers

Use
---
    This module can be executed in a Python shell as such:

    ::
        from mirage.catalogs import get_catalog
        sw_cats, lw_cats = get_catalog.get_all_catalogs(pointing_file, prop_id)

    Required arguments:
        ``pointing file`` - filepath to the APT pointing file
        ``prop_id`` - APT proposal ID number

References
----------
    Example of a query with Vizier:
        http://astroquery.readthedocs.io/en/latest/vizier/vizier.html
'''
import os

from astropy.io import ascii as asc
from astropy import units as u
from astropy.coordinates import SkyCoord
from astroquery.vizier import Vizier

from ..apt import apt_inputs

SCRIPTS_DIR = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
PACKAGE_DIR = os.path.dirname(SCRIPTS_DIR)
CATALOG_DIR = os.path.join(PACKAGE_DIR, 'catalogs')


def get_target_coords(pointing_file, prop_id):
    '''Parse APT file to extract RA and Dec from each observation

    Parameters
    ----------
    pointing_file : str
        Path to APT .pointing file
    prop_id : str
        APT proposal ID number

    Returns
    -------
    target_coords : list
        List of astropy SkyCoord objects for all targets
    '''
    APT = apt_inputs.AptInput()
    pointing = APT.get_pointing_info(pointing_file, prop_id)

    # Split up by observation, finding all RA and Decs in pointing file
    obs_nums = []
    coords = []
    for ra, dec, obs_id in zip(pointing['ra'], pointing['dec'], pointing['obs_num']):
        if obs_id not in obs_nums:
            obs_nums.append(obs_id)
            coords.append((ra, dec))
        else:
            # If there are different RA/Decs within an observation.... eek
            if (ra != coords[-1][0]) or (dec != coords[-1][1]):
                print('** WARNING: Different RA/Decs within observation {} **'.format(obs_id))

    target_coords = []
    for ra, dec in coords:
        target_coords.append(SkyCoord(ra, dec, unit=(u.deg, u.deg)))

    for tup in set([(t.ra.deg, t.dec.deg) for t in target_coords]):
        print('Target coordinates: ', tup[0], tup[1])

    return target_coords


def get_sw_catalog(target_coords, search_radius=15 * u.arcmin):
    '''Query 2MASS catalog and write out shortwave catalog file for each target

    Parameters
    ----------
    target_coords : list
        List of astropy SkyCoord objects for all targets

    Returns
    -------
    catalog_filenames_sw : list
        List of filenames for all shortwave catalogs

    References
    ----------
    Catalog: http://vizier.u-strasbg.fr/viz-bin/VizieR?-source=II/246
    '''
    catalog_filenames_sw = []

    # For each target
    for i_obs, t in enumerate(target_coords):
        # Generate the catalog filename
        catalog_filename_sw = os.path.join(CATALOG_DIR,
                                           '2MASS_RA{:.2f}deg_Dec{:.2f}deg.list'.format(t.ra.degree,
                                                                                        t.dec.degree))

        # Check that the shortwave catalog has not already been created
        if os.path.isfile(catalog_filename_sw):
            catalog_filenames_sw.append(catalog_filename_sw)
            if catalog_filename_sw not in catalog_filenames_sw[:i_obs]:
                print('Shortwave catalog file {} already exists. Will not overwrite.'.\
                      format(os.path.basename(catalog_filename_sw)))
            continue

        # If not, query shortwave sources from the 2MASS catalog from Vizier
        v = Vizier(catalog='II/246/out', columns=['_RAJ2000', '_DEJ2000', 'Kmag'])
        v.ROW_LIMIT = -1
        result = v.query_region(t, radius=search_radius)
        queried_catalog_sw = result['II/246/out']

        print('Queried {} 2MASS objects within {} {} of RA, Dec ({:.2f}, {:.2f}).'.
              format(len(queried_catalog_sw), search_radius.value,
                     str(search_radius.unit), t.ra.deg, t.dec.deg))

        # Rename table columns to match mirage's expectations
        queried_catalog_sw.rename_column('_RAJ2000', 'x_or_RA')
        queried_catalog_sw.rename_column('_DEJ2000', 'y_or_Dec')
        queried_catalog_sw.rename_column('Kmag', 'magnitude')

        # Save shortwave catalog and header comments to file
        comments = [
            '',
            'vegamag',
            'Catalog for ramp_simulator.py. created from 2MASS catalog',
            'Matching NIRCam SW F212N filter',
            'Magnitudes are Kmag (2-3 microns).'
        ]
        queried_catalog_sw.meta['comments'] = comments
        asc.write(queried_catalog_sw, output=catalog_filename_sw, overwrite=True)
        print('Successfully saved shortwave catalog to {}.'.format(catalog_filename_sw))

        catalog_filenames_sw.append(catalog_filename_sw)

    return catalog_filenames_sw


def get_lw_catalog(target_coords, search_radius=15 * u.arcmin):
    '''Query WISE catalog and write out longwave catalog file for each target

    Parameters
    ----------
    target_coords : list
        List of astropy SkyCoord objects for all targets

    Returns
    -------
    catalog_filenames_lw : list
        List of filenames for all longwave catalogs

    References
    ----------
    Catalog: http://vizier.u-strasbg.fr/viz-bin/VizieR?-source=II/328
    '''
    catalog_filenames_lw = []

    # For each target
    for i_obs, t in enumerate(target_coords):
        # Generate the catalog filename
        catalog_filename_lw = os.path.join(CATALOG_DIR,
                                           'WISE_RA{:.2f}deg_Dec{:.2f}deg.list'.format(t.ra.degree,
                                                                                       t.dec.degree))

        # Check that the shortwave catalog has not already been created
        if os.path.isfile(catalog_filename_lw):
            catalog_filenames_lw.append(catalog_filename_lw)
            if catalog_filename_lw not in catalog_filenames_lw[:i_obs]:
                print('Longwave catalog file {} already exists. Will not overwrite.'.\
                      format(os.path.basename(catalog_filename_lw)))
            continue

        # If not, query longwave sources from the WISE catalog from Vizier
        v = Vizier(catalog='II/328/allwise', columns=['RAJ2000', 'DEJ2000', 'W2mag'])
        v.ROW_LIMIT = -1
        result = v.query_region(t, radius=search_radius)
        queried_catalog_lw = result['II/328/allwise']

        print('Queried {} WISE objects within {} {} of RA, Dec ({:.2f}, {:.2f}).'.
              format(len(queried_catalog_lw), search_radius.value,
                     str(search_radius.unit), t.ra.deg, t.dec.deg))

        # Rename table columns to match nircam simulator expectations
        queried_catalog_lw.rename_column('RAJ2000', 'x_or_RA')
        queried_catalog_lw.rename_column('DEJ2000', 'y_or_Dec')
        queried_catalog_lw.rename_column('W2mag', 'magnitude')

        # Save longwave catalog and header comments to file
        comments = [
            '',
            'abmag',
            'Catalog for ramp_simulator.py. created from WISE catalog',
            'Matching NIRCam LW F460M & F480M filters',
            'Magnitudes are W2 (4-8 microns).'
        ]
        queried_catalog_lw.meta['comments'] = comments
        asc.write(queried_catalog_lw, output=catalog_filename_lw, overwrite=True)
        print('Successfully saved longwave catalog to {}.'.format(catalog_filename_lw))

        catalog_filenames_lw.append(catalog_filename_lw)

    return catalog_filenames_lw


def get_all_catalogs(pointing_file, prop_id):
    '''Query WISE and 2MASS catalogs and write out long- and shortwave
    catalog files for each target in the provided APT proposal

    Parameters
    ----------
    pointing_file : str
        Path to APT .pointing file
    prop_id : str
        APT proposal ID number

    Returns
    -------
    tuple
        Lists of shortwave and longwave catalog filenames
    '''
    target_coords = get_target_coords(pointing_file, prop_id)
    catalog_filenames_sw = get_sw_catalog(target_coords)
    catalog_filenames_lw = get_lw_catalog(target_coords)

    return target_coords, catalog_filenames_sw, catalog_filenames_lw



# ! /usr/bin/env python

'''
Given APT output files, read in data relevant to the data simulator,
organize, and create input files for the simulator.

Inputs:

xml file - Name of xml file exported from APT.
pointing file - Name of associated pointing file exported from APT.

Optional inputs:

epoch_list - Name of ascii file which lists observation labels and
             associated starting observation time, as well as telescope
             roll angle (PAV3). If you wish to have observations back-
             to-back, give them all the same starting time.

Outputs:

output_csv - Ascii table containing all relevant information needed
             to construct a ramp simulator input file for each
             observation/dither/detector combination.

Dependencies:

argparse, lxml, astropy, numpy, collections

JWST Calibration pipeline (only for the use of set_telescope_pointing.py)
in order to translate PAV3 values to local roll angles for each detector.


HISTORY:

July 2017 - V0: Initial version. Bryan Hilbert
Feb 2018  - V1: Updated to work for multiple filter pairs per observation
            Lauren Chambers
August 2018 - V2: Replaced manual Ra, Dec calculations with pysiaf functionality
October 2018 - Major modifications to read programs of all science instruments and parallels
               Johannes Sahlmann
'''
import copy
import os
import re
import argparse
import pkg_resources

from astropy.table import Table, vstack
from astropy.io import ascii
import numpy as np
from pysiaf import rotations
import yaml

from . import read_apt_xml
from ..utils import siaf_interface, constants, utils


class AptInput:
    """Summary

    Attributes:
        exposure_tab (TYPE): Description
        input_xml (str): Description
        observation_list_file (str): Description
        obstab (TYPE): Description
        output_csv (TYPE): Description
        pointing_file (str): Description
    """

    def __init__(self, input_xml=None, pointing_file=None, output_dir=None, output_csv=None,
                 observation_list_file=None):
        self.input_xml = input_xml
        self.pointing_file = pointing_file
        self.output_dir = output_dir
        self.output_csv = output_csv
        self.observation_list_file = observation_list_file

        # Locate the module files, so that we know where to look
        # for config subdirectory
        self.config_path = os.path.join(pkg_resources.resource_filename('mirage', ''), 'config')

    def add_epochs(self, intab):
        """NOT CURRENTLY USED"""
        # add information on the epoch of each observation
        # if the user entered a list of epochs, read that in
        default_date = '2020-10-14'

        if self.epoch_list is not None:
            epochs = ascii.read(self.epoch_list, header_start=0, data_start=1)
        else:
            epochs = Table()
            epochs['observation'] = intab['obs_label']
            epochs['date'] = ['2018-10-14'] * len(intab['obs_label'])
            epochs['pav3'] = [0.] * len(intab['obs_label'])

        # insert epoch info for each observation into dictionary
        epoch_start = []
        epoch_pav3 = []
        for obs in intab['obs_label']:
            match = obs == epochs['observation'].data
            if np.sum(match) == 0:
                print("No valid epoch line found for observation {}".format(obs))
                print(epochs['observation'].data)
                epoch_start.append(default_date)
                epoch_pav3.append(0.)
            else:
                epoch_start.append(epochs['date'][match].data[0])
                epoch_pav3.append(epochs['pav3'][match].data[0])
        intab['epoch_start_date'] = epoch_start
        intab['pav3'] = epoch_pav3
        return intab

    def add_observation_info(self, intab):
        """Add information about each observation.

        Catalog names, dates, PAV3 values, etc., which are retrieved from the observation list
        yaml file.

        Parameters
        ----------
        intab : obj
            astropy.table.Table containing exposure information

        Returns
        -------
        intab : obj
            Updated table with information from the observation list
            yaml file added.

        """
        with open(self.observation_list_file, 'r') as infile:
            self.obstab = yaml.safe_load(infile)

        OBSERVATION_LIST_FIELDS = 'Date PAV3 Filter PointSourceCatalog GalaxyCatalog ' \
                                  'ExtendedCatalog ExtendedScale ExtendedCenter MovingTargetList ' \
                                  'MovingTargetSersic MovingTargetExtended ' \
                                  'MovingTargetConvolveExtended MovingTargetToTrack ' \
                                  'ImagingTSOCatalog GrismTSOCatalog ' \
                                  'BackgroundRate DitherIndex CosmicRayLibrary CosmicRayScale'.split()

        nircam_mapping = {'ptsrc': 'PointSourceCatalog',
                          'galcat': 'GalaxyCatalog',
                          'ext': 'ExtendedCatalog',
                          'extscl': 'ExtendedScale',
                          'extcent': 'ExtendedCenter',
                          'movptsrc': 'MovingTargetList',
                          'movgal': 'MovingTargetSersic',
                          'movext': 'MovingTargetExtended',
                          'movconv': 'MovingTargetConvolveExtended',
                          'solarsys': 'MovingTargetToTrack',
                          'img_tso': 'ImagingTSOCatalog',
                          'grism_tso': 'GrismTSOCatalog',
                          'bkgd': 'BackgroundRate',
                          }

        unique_instrument_names = [name.lower() for name in np.unique(intab['Instrument'])]

        # initialize dictionary keys
        for key in OBSERVATION_LIST_FIELDS:
            intab[key] = []

        if 'nircam' in unique_instrument_names:
            for channel in ['SW', 'LW']:
                for name, item in nircam_mapping.items():
                    key = '{}_{}'.format(channel.lower(), name)
                    intab[key] = []

        # loop over entries in input dictionary
        for index, instrument in enumerate(intab['Instrument']):
            instrument = instrument.lower()

            # retrieve corresponding entry from observation list
            entry = get_entry(self.obstab, intab['entry_number'][index])

            if instrument == 'nircam':
                # keep the number of entries in the dictionary consistent
                for key in OBSERVATION_LIST_FIELDS:
                    if key in ['Date', 'PAV3', 'Instrument', 'CosmicRayLibrary', 'CosmicRayScale']:
                        value = str(entry[key])
                    else:
                        value = str(None)

                    intab[key].append(value)

                for channel in ['SW', 'LW']:
                    for name, item in nircam_mapping.items():
                        key = '{}_{}'.format(channel.lower(), name)
                        if item in 'ExtendedScale ExtendedCenter MovingTargetConvolveExtended BackgroundRate'.split():
                            intab[key].append(entry['FilterConfig'][channel][item])
                        else:
                            intab[key].append(self.full_path(entry['FilterConfig'][channel][item]))

            else:
                for key in OBSERVATION_LIST_FIELDS:
                    value = str(entry[key])

                    # Expand catalog names to contain full paths
                    catalog_names = 'PointSourceCatalog GalaxyCatalog ' \
                                    'ExtendedCatalog MovingTargetList ' \
                                    'MovingTargetSersic MovingTargetExtended ' \
                                    'MovingTargetToTrack ImagingTSOCatalog ' \
                                    'GrismTSOCatalog'.split()
                    if key in catalog_names:
                        value = self.full_path(value)
                    intab[key].append(value)

                # keep the number of entries in the dictionary consistent
                if 'nircam' in unique_instrument_names:
                    for channel in ['SW', 'LW']:
                        for name, item in nircam_mapping.items():
                            key = '{}_{}'.format(channel.lower(), name)
                            intab[key].append(str(None))

        intab['epoch_start_date'] = intab['Date']
        return intab

    def base36encode(self, integer):
        """
        Translate a base 10 integer to base 36

        Parameters
        ----------
        integer : int
            a base 10 integer

        Returns
        -------
        integer : int
            The integer translated to base 36
        """
        chars, encoded = '0123456789abcdefghijklmnopqrstuvwxyz', ''

        while integer > 0:
            integer, remainder = divmod(integer, 36)
            encoded = chars[remainder] + encoded

        return encoded.zfill(2)

    def combine_dicts(self, dict1, dict2):
        """Combine two dictionaries into a single dictionary.

        Parameters
        ----------
        dict1 : dict
            dictionary
        dict2 : dict
            dictionary

        Returns
        -------
        combined : dict
            Combined dictionary
        """
        combined = dict1.copy()
        combined.update(dict2)
        return combined

    def create_input_table(self, verbose=False):
        """

        Expansion for dithers is done upstream.

        Parameters
        ----------
        verbose

        Returns
        -------

        """
        # Expand paths to full paths
        # self.input_xml = os.path.abspath(self.input_xml)
        # self.pointing_file = os.path.abspath(self.pointing_file)
        if self.output_csv is not None:
            self.output_csv = os.path.abspath(self.output_csv)
        if self.observation_list_file is not None:
            self.observation_list_file = os.path.abspath(self.observation_list_file)

        # main_dir = os.path.split(self.input_xml)[0]

        # if APT.xml content has already been generated during observation list creation
        # (generate_observationlist.py) load it here

        if self.apt_xml_dict is None:
            raise RuntimeError('self.apt_xml_dict is not defined')

        # Read in the pointing file and produce dictionary
        pointing_dictionary = self.get_pointing_info(self.pointing_file, propid=self.apt_xml_dict['ProposalID'][0])

        # Check that the .xml and .pointing files agree
        assert len(self.apt_xml_dict['ProposalID']) == len(pointing_dictionary['obs_num']),\
            ('Inconsistent table size from XML file ({}) and pointing file ({}). Something was not '
             'processed correctly in apt_inputs.'.format(len(self.apt_xml_dict['ProposalID']),
                                                         len(pointing_dictionary['obs_num'])))

        # Combine the dictionaries
        observation_dictionary = self.combine_dicts(self.apt_xml_dict, pointing_dictionary)

        # Add epoch and catalog information
        observation_dictionary = self.add_observation_info(observation_dictionary)

        # if verbose:
        #     print('Summary of observation dictionary:')
        #     for key in observation_dictionary.keys():
        #         print('{:<25}: number of elements is {:>5}'.format(key, len(observation_dictionary[key])))

        self.exposure_tab = self.expand_for_detectors(observation_dictionary)

        # fix data for filename generation
        # set parallel seq id
        for j, isparallel in enumerate(self.exposure_tab['ParallelInstrument']):
            if isparallel:
                self.exposure_tab['sequence_id'][j] = '2'

        # set exposure number (new sequence for every combination of seq id and act id and observation number and detector)
        temp_table = Table([self.exposure_tab['sequence_id'], self.exposure_tab['exposure'], self.exposure_tab['act_id'], self.exposure_tab['obs_num'], self.exposure_tab['detector']], names=('sequence_id', 'exposure', 'act_id', 'obs_num', 'detector'))

        for obs_num in np.unique(self.exposure_tab['obs_num']):
            for act_id in np.unique(temp_table['act_id']):
                # prime_index = np.where((temp_table['sequence_id'] == '1') & (temp_table['act_id'] == act_id) & (temp_table['obs_num'] == obs_num))[0]
                # parallel_index = np.where((temp_table['sequence_id'] == '2') & (temp_table['act_id'] == act_id) & (temp_table['obs_num'] == obs_num))[0]
                for detector in np.unique(temp_table['detector']):
                    prime_index = np.where((temp_table['sequence_id'] == '1') & (temp_table['act_id'] == act_id) & (temp_table['obs_num'] == obs_num) & (temp_table['detector']==detector))[0]
                    parallel_index = np.where((temp_table['sequence_id'] == '2') & (temp_table['act_id'] == act_id) & (temp_table['obs_num'] == obs_num) & (temp_table['detector']==detector))[0]

                    temp_table['exposure'][prime_index] = ['{:05d}'.format(n+1) for n in np.arange(len(prime_index))]
                    temp_table['exposure'][parallel_index] = ['{:05d}'.format(n+1) for n in np.arange(len(parallel_index))]
        self.exposure_tab['exposure'] = list(temp_table['exposure'])

        self.check_aperture_override()

        if verbose:
            for key in self.exposure_tab.keys():
                print('{:>20} has {:>10} items'.format(key, len(self.exposure_tab[key])))

        # Create a pysiaf.Siaf instance for each instrument in the proposal
        self.siaf = {}
        for instrument_name in np.unique(observation_dictionary['Instrument']):
            self.siaf[instrument_name] = siaf_interface.get_instance(instrument_name)

        # Calculate the correct V2, V3 and RA, Dec for each exposure/detector
        self.ra_dec_update()

        # Output to a csv file.
        if self.output_csv is None:
            indir, infile = os.path.split(self.input_xml)
            self.output_csv = os.path.join(self.output_dir, 'Observation_table_for_' + infile.split('.')[0] + '.csv')
        ascii.write(Table(self.exposure_tab), self.output_csv, format='csv', overwrite=True)
        print('csv exposure list written to {}'.format(self.output_csv))

    def check_aperture_override(self):
        if bool(self.exposure_tab['FiducialPointOverride']) is True:
            instruments = self.exposure_tab['Instrument']
            apertures = self.exposure_tab['aperture']

            aperture_key = constants.instrument_abbreviations

            fixed_apertures = []
            for i, (instrument, aperture) in enumerate(zip(instruments, apertures)):
                inst_match_ap = aperture.startswith(aperture_key[instrument.lower()])
                if not inst_match_ap:
                    # Handle the one case we understand, for now
                    if instrument.lower() == 'fgs' and aperture[:3] == 'NRC':
                        obs_num = self.exposure_tab['obs_num'][i]
                        guider_number = read_apt_xml.get_guider_number(self.input_xml, obs_num)
                        guider_aperture = 'FGS{}_FULL'.format(guider_number)
                        fixed_apertures.append(guider_aperture)
                    else:
                        print(instrument, aperture, inst_match_ap, aperture_key[instrument.lower()])
                        raise ValueError('Unknown FiducialPointOverride in program. Instrument = {} but aperture = {}.'.format(instrument, aperture))
                else:
                    fixed_apertures.append(aperture)

            # Add new dictionary entry to document the FiducialPointOverride (pointing aperture)
            self.exposure_tab['pointing_aperture'] = self.exposure_tab['aperture']
            # Rewrite the existing imaging aperture entry to match the primary instrument
            self.exposure_tab['aperture'] = fixed_apertures
        else:
            # Add new dictionary entry to document that the imaging aperture is the
            # same as the pointing aperture
            self.exposure_tab['pointing_aperture'] = self.exposure_tab['aperture']

    def expand_for_detectors(self, input_dictionary):
        """Expand dictionary to have one entry per detector, rather than the
        one line per module that is in the input

        Parameters
        ----------
        input_dictionary : dict
            dictionary containing one entry per module

        Returns
        -------
        observation_dictionary : dict
            dictionary expanded to have one entry per detector
        """
        observation_dictionary = {}
        for key in input_dictionary:
            observation_dictionary[key] = []
        observation_dictionary['detector'] = []

        # Read in tables of aperture information
        nircam_subarray_file = os.path.join(self.config_path, 'NIRCam_subarray_definitions.list')
        nircam_apertures = utils.read_subarray_definition_file(nircam_subarray_file)

        for index, instrument in enumerate(input_dictionary['Instrument']):
            instrument = instrument.lower()
            if instrument == 'nircam':
                # NIRCam case: Expand for detectors. Create one entry in each list for each
                # detector, rather than a single entry for 'ALL' or 'BSALL'

                # Determine module and subarray of the observation
                sub = input_dictionary['Subarray'][index]
                module = input_dictionary['Module'][index]
                if module == 'ALL':
                    module = ['A', 'B']
                else:
                    module = list(module)

                # Match up `sub` with aperture names in the aperture table
                # FULL matches up with the standard full frame imaging
                # apertures as well as full frame Grism apertures
                matches = np.where(nircam_apertures['Name'] == sub)[0]

                if len(matches) == 0:
                    raise ValueError("ERROR: aperture {} in not present in the subarray definition file {}"
                                     .format(sub, nircam_subarray_file))

                # Keep only apertures in the correct module
                matched_allmod_apertures = nircam_apertures['AperName'][matches].data
                matched_apertures = []
                for mod in module:
                    good = [ap for ap in matched_allmod_apertures if 'NRC{}'.format(mod) in ap]
                    matched_apertures.extend(good)

                if sub in ['FULL', 'SUB160', 'SUB320', 'SUB640', 'SUB64P', 'SUB160P', 'SUB400P', 'FULLP']:
                    mode = input_dictionary['Mode'][index]
                    if (sub == 'FULL'):

                        if mode in ['imaging', 'ts_imaging', 'wfss']:
                            # This block should catch full-frame observations
                            # in either imaging (including TS imaging) or
                            # wfss mode
                            matched_aps = np.array([ap for ap in matched_apertures if 'GRISM' not in ap])
                            matched_apertures = []
                            detectors = []
                            for ap in matched_aps:
                                detectors.append(ap[3:5])
                                split = ap.split('_')
                                if len(split) == 3:
                                    ap_string = '{}_{}'.format(split[1], split[2])
                                elif len(split) == 2:
                                    ap_string = split[1]
                                matched_apertures.append(ap_string)

                        elif mode == 'ts_grism':
                            # This block should get Grism Time Series
                            # observations that use the full frame
                            matched_apertures = np.array([ap for ap in matched_apertures if 'GRISM' in ap])
                            filtered_aperture = input_dictionary['aperture'][index]
                            filtered_splits = filtered_aperture.split('_')
                            filtered_ap_no_det = '{}_{}'.format(filtered_splits[1], filtered_splits[2])
                            detectors = [filtered_splits[0][3:5]]

                            # Get correct apertures
                            apertures_to_add = []

                            final_matched_apertures = []
                            final_detectors = []
                            filtered_ap_det = filtered_aperture[3:5]
                            for ap in matched_apertures:
                                det = ap[3:5]
                                if ap == filtered_aperture or det != filtered_ap_det:
                                    final_matched_apertures.append(ap)
                                    final_detectors.append(det)
                            matched_apertures = final_matched_apertures
                            detectors = final_detectors
                    else:
                        # 'Standard' imaging subarrays: SUB320, SUB400P, etc
                        matched_apertures = [ap for ap in matched_apertures if sub in ap]
                        detectors = [ap.split('_')[0][3:5] for ap in matched_apertures]
                        matched_apertures = [sub] * len(detectors)

                elif 'SUBGRISM' in sub:
                    # This should catch only Grism Time Series observations
                    # and engineering imaging observations, which are the
                    # only 2 templates that can use SUBGRISM apertures
                    long_filter = input_dictionary['LongFilter'][index]
                    filter_dependent_apertures = [ap for ap in matched_apertures if len(ap.split('_')) == 3]

                    filtered_aperture = input_dictionary['aperture'][index]
                    filtered_splits = filtered_aperture.split('_')
                    filtered_ap_no_det = '{}_{}'.format(filtered_splits[1], filtered_splits[2])
                    detectors = [filtered_splits[0][3:5]]

                    # Get correct apertures
                    apertures_to_add = []

                    final_matched_apertures = []
                    final_detectors = []
                    filtered_ap_det = filtered_aperture[3:5]
                    for ap in matched_apertures:
                        det = ap[3:5]
                        if ap == filtered_aperture or det != filtered_ap_det:
                            final_matched_apertures.append(ap)
                            final_detectors.append(det)
                    matched_apertures = []
                    for ap in final_matched_apertures:
                        split = ap.split('_')
                        if len(split) == 3:
                            ap_string = '{}_{}'.format(split[1], split[2])
                        elif len(split) == 2:
                            ap_string = split[1]
                        matched_apertures.append(ap_string)
                    detectors = final_detectors
                else:
                    # TA, WFSC apertures
                    stripped_apertures = []
                    detectors = []
                    for ap in matched_apertures:
                        detectors.append(ap[3:5])
                        split = ap.split('_')
                        if len(split) == 3:
                            ap_string = '{}_{}'.format(split[1], split[2])
                        elif len(split) == 2:
                            ap_string = split[1]
                        stripped_apertures.append(ap_string)
                    matched_apertures = stripped_apertures

                full_apertures = ['NRC{}_{}'.format(det, sub) for det, sub in zip(detectors, matched_apertures)]

                # Add entries to observation dictionary
                num_entries = len(detectors)
                #observation_dictionary['Subarray'].extend(matched_apertures) extend? or replace?
                for key in input_dictionary:
                    #if key not in ['Subarray']:
                    if key not in ['aperture', 'detector', 'Subarray']:
                        observation_dictionary[key].extend(([input_dictionary[key][index]] * num_entries))
                observation_dictionary['detector'].extend(detectors)
                observation_dictionary['aperture'].extend(full_apertures)
                observation_dictionary['Subarray'].extend(matched_apertures)

            else:
                if instrument == 'niriss':
                    detectors = ['NIS']

                elif instrument == 'nirspec':
                    detectors = ['NRS']

                elif instrument == 'fgs':
                    guider_number = read_apt_xml.get_guider_number(self.input_xml, input_dictionary['obs_num'][index])
                    detectors = ['G{}'.format(guider_number)]

                elif instrument == 'miri':
                    detectors = ['MIR']

                n_detectors = len(detectors)
                for key in input_dictionary:
                    observation_dictionary[key].extend(([input_dictionary[key][index]] * n_detectors))
                observation_dictionary['detector'].extend(detectors)

        """
        # Correct NIRCam aperture names for commissioning subarrays
        for index, instrument in enumerate(observation_dictionary['Instrument']):
            instrument = instrument.lower()
            if instrument == 'nircam':
                detector = observation_dictionary['detector'][index]
                sub = observation_dictionary['Subarray'][index]

                # this should probably better be handled by using the subarray_definitions file upstream
                if 'DHSPIL' in sub:
                    subarray, module = sub.split('DHSPIL')
                    subarray_size = subarray[3:]
                    detector = 'A3' if module == 'A' else 'B4'
                    aperture_name = 'NRC{}_DHSPIL_SUB{}'.format(detector, subarray_size)
                elif 'FP1' in sub:
                    subarray, module = sub.split('FP1')
                    subarray_size = subarray[3:]
                    detector = 'A3' if module == 'A' else 'B4'
                    aperture_name = 'NRC{}_FP1_SUB{}'.format(detector, subarray_size)
                else:
                    aperture_name = 'NRC' + detector + '_' + sub

                observation_dictionary['aperture'][index] = aperture_name
                observation_dictionary['detector'][index] = detector
        """

        return observation_dictionary

    def extract_grism_aperture(self, apertures, filter_name):
        """In the case of a Grism observation (WFSS or GRISM TSO), where a
        given crossing filter is used, find the appropriate aperture to
        use.

        Paramters
        ---------
        apertures : list
            List of possible grism apertures

        filter_name : str
            Name of crossing filter

        Returns
        -------
        apertures : list
            Modified list containig the correct aperture
        """
        filter_match = [True if filter_name in mtch else False for mtch in apertures]
        if any(filter_match):
            print('EXACT FILTER MATCH')
            print(filter_match)
            apertures = list(np.array(apertures)[filter_match])
        else:
            print('NO EXACT FILTER MATCH')
            filter_int = int(filter_name[1:4])
            aperture_int = np.array([int(ap.split('_')[-1][1:4]) for ap in apertures])
            wave_diffs = np.abs(aperture_int - filter_int)
            min_diff_index = np.where(wave_diffs == np.min(wave_diffs))[0]
            apertures = list(apertures[min_diff_index])

            print(filter_int, aperture_int, min_diff_index, apertures)

        return apertures

    def extract_value(self, line):
        """Extract text from xml line

        Parameters
        ----------
        line : str
            Line from xml file

        Returns
        -------
        line : str
            Text between > and < in the input line
        """
        gt = line.find('>')
        lt = line.find('<', gt)
        return line[gt + 1:lt]

    def full_path(self, in_path):
        """
        If the input path is not None, expand
        any environment variables and make an
        absolute path. Return the updated path.

        Parameters
        ----------
        in_path : str
            Path to be expanded

        Returns
        -------
        in_path : str
            Expanded, absolute path
        """
        if in_path.lower() == 'none':
            return in_path
        else:
            return os.path.abspath(os.path.expandvars(in_path))

    def get_pointing_info(self, file, propid=0, verbose=False):
        """Read in information from APT's pointing file.

        Parameters
        ----------
        file : str
            Name of APT-exported pointing file to be read
        propid : int
            Proposal ID number (integer). This is used to
            create various ID fields

        Returns
        -------
        pointing : dict
            Dictionary of pointing-related information

        TODO
        ----
            extract useful information from header?
            check visit numbers
            set parallel proposal number correctly

        """
        tar = []
        tile = []
        exp = []
        dith = []
        aperture = []
        targ1 = []
        targ2 = []
        ra = []
        dec = []
        basex = []
        basey = []
        dithx = []
        dithy = []
        v2 = []
        v3 = []
        idlx = []
        idly = []
        level = []
        type_str = []
        expar = []
        dkpar = []
        ddist = []
        observation_number = []
        visit_number = []
        visit_id = []
        visit_grp = []
        activity_id = []
        observation_label = []
        observation_id = []
        seq_id = []

        act_counter = 1
        with open(file) as f:
            for line in f:

                # skip comments and new lines
                if (line[0] == '#') or (line in ['\n']) or ('=====' in line):
                    continue
                # extract proposal ID
                elif line.split()[0] == 'JWST':
                    propid_header = line.split()[7]
                    try:
                        propid = np.int(propid_header)
                    except ValueError:
                        # adopt value passed to function
                        pass
                    if verbose:
                        print('Extracted proposal ID {}'.format(propid))
                    continue

                elif (len(line) > 1):
                    elements = line.split()

                    # Look for lines that give visit/observation numbers
                    if line[0:2] == '* ':
                        paren = line.rfind('(')
                        if paren == -1:
                            obslabel = line[2:]
                            obslabel = obslabel.strip()
                        else:
                            obslabel = line[2:paren-1]
                            obslabel = obslabel.strip()
                        if (' (' in obslabel) and (')' in obslabel):
                            obslabel = re.split(r' \(|\)', obslabel)[0]

                    skip = False

                    if line[0:2] == '**':
                        v = elements[2]
                        obsnum, visitnum = v.split(':')
                        obsnum = str(obsnum).zfill(3)
                        visitnum = str(visitnum).zfill(3)
                        if (skip is True) and (verbose):
                            print('Skipping observation {} ({})'.format(obsnum, obslabel))

                    try:
                        # Skip the line at the beginning of each
                        # visit that gives info on the target,
                        # but is not actually an observation
                        # These lines have 'Exp' values of 0,
                        # while observations have a value of 1
                        # (that I've seen so far)

                        # The exception to this rule is TA images.
                        # These have Exp values of 0. Sigh.


                        if ((np.int(elements[1]) > 0) & ('NRC' in elements[4]
                                                         or 'NIS' in elements[4]
                                                         or 'FGS' in elements[4]
                                                         or 'NRS' in elements[4]
                                                         or 'MIR' in elements[4])
                            ) or (('TA' in elements[4]) & ('NRC' in elements[4]
                                                         or 'NIS' in elements[4]
                                                         or 'FGS' in elements[4]
                                                         or 'NRS' in elements[4]
                                                         or 'MIR' in elements[4])):
                            if (elements[18] == 'PARALLEL') and ('MIRI' in elements[4]):
                                skip = True

                            if skip:
                                # act_counter += 1
                                continue
                            act = self.base36encode(act_counter)
                            activity_id.append(act)
                            observation_label.append(obslabel)
                            observation_number.append(obsnum)
                            visit_number.append(visitnum)
                            prop_5digit = "{0:05d}".format(int(propid))
                            vid = "{}{}{}".format(prop_5digit, obsnum, visitnum)
                            visit_id.append(vid)
                            # Visit group hard coded to 1. It's not clear how APT divides visits up into visit
                            # groups. For now just keep everything in a single visit group.
                            vgrp = '01'
                            visit_grp.append(vgrp)
                            # Parallel sequence id hard coded to 1 (Simulated instrument as prime rather than
                            # parallel) at the moment. Future improvements may allow the proper sequence
                            # number to be constructed.
                            seq = '1'
                            seq_id.append(seq)
                            tar.append(np.int(elements[0]))
                            tile.append(np.int(elements[1]))
                            exnum = str(elements[2]).zfill(5)
                            exp.append(exnum)
                            dith.append(np.int(elements[3]))

                            ap = elements[4]
                            if ('GRISMR_WFSS' in elements[4]):
                                ap = ap.replace('GRISMR_WFSS', 'FULL')
                            elif ('GRISMC_WFSS' in elements[4]):
                                ap = ap.replace('GRISMC_WFSS', 'FULL')

                            aperture.append(ap)
                            targ1.append(np.int(elements[5]))
                            targ2.append(elements[6])
                            ra.append(elements[7])
                            dec.append(elements[8])
                            basex.append(elements[9])
                            basey.append(elements[10])
                            dithx.append(np.float(elements[11]))
                            dithy.append(np.float(elements[12]))
                            v2.append(np.float(elements[13]))
                            v3.append(np.float(elements[14]))
                            idlx.append(np.float(elements[15]))
                            idly.append(np.float(elements[16]))
                            level.append(elements[17])
                            type_str.append(elements[18])
                            expar.append(np.int(elements[19]))
                            dkpar.append(np.int(elements[20]))
                            #if elements[18] == 'PARALLEL':
                            #    ddist.append(None)
                            #else:
                                #print('line is: {}'.format(elements))
                                #print(ddist)
                                #ddist.append(np.float(elements[21]))
                            # For the moment we assume that the instrument being simulated is not being
                            # run in parallel, so the parallel proposal number will be all zeros,
                            # as seen in the line below.
                            observation_id.append("V{}P{}{}{}{}".format(vid, '00000000', vgrp, seq, act))
                            # act_counter += 1

                    except ValueError as e:
                        if verbose:
                            print('Skipping line:\n{}\nproducing error:\n{}'.format(line, e))
                        pass

        pointing = {'exposure': exp, 'dither': dith, 'aperture': aperture,
                    'targ1': targ1, 'targ2': targ2, 'ra': ra, 'dec': dec,
                    'basex': basex, 'basey': basey, 'dithx': dithx,
                    'dithy': dithy, 'v2': v2, 'v3': v3, 'idlx': idlx,
                    'idly': idly, 'obs_label': observation_label,
                    'obs_num': observation_number, 'visit_num': visit_number,
                    'act_id': activity_id, 'visit_id': visit_id, 'visit_group': visit_grp,
                    'sequence_id': seq_id, 'observation_id': observation_id}
        return pointing

    def tight_dithers(self, input_dict):
        """
        In NIRCam, when the 'FULL' dither pattern is
        used, it is possible to set the number of primary
        dithers to '3TIGHT' rather than just a number
        (e.g. '3'). If the number of dithers is set to '3TIGHT'
        remove 'TIGHT' from the entries and leave
        only the number behind.

        Parameters
        ----------

        input_dict : dict
           Dictionary where each key points to a list containing
           observation details. For example, input_dict['PrimarDither']
           is a list of the number of primary dithers for all observations

        Returns
        -------

        input_dict : dict
            Updated dictionary where 'TIGHT' has been removed from
            PrimaryDither list
        """
        inlist = input_dict['PrimaryDithers']
        modlist = [v if 'TIGHT' not in v else v.strip('TIGHT') for v in inlist]
        input_dict['PrimaryDithers'] = modlist
        return input_dict

    def ra_dec_update(self, verbose=False):
        """Given the V2, V3 values for the reference pixels associated
        with detector apertures, calculate corresponding RA, Dec.
        """
        sw_grismts_apertures = ['NRCA1_GRISMTS256', 'NRCA1_GRISMTS128', 'NRCA1_GRISMTS64',
                                'NRCA3_GRISMTS256', 'NRCA3_GRISMTS128', 'NRCA3_GRISMTS64']

        lw_grismts_apertures = ['NRCA5_GRISM256_F322W2', 'NRCA5_GRISM128_F322W2', 'NRCA5_GRISM64_F322W2',
                                'NRCA5_GRISM256_F444W', 'NRCA5_GRISM128_F444W', 'NRCA5_GRISM64_F444W']

        intermediate_lw_grismts_apertures = ['NRCA5_TAGRISMTS_SCI_F444W', 'NRCA5_TAGRISMTS_SCI_F322W2']

        aperture_ra = []
        aperture_dec = []

        lw_grismts_aperture = None
        for i in range(len(self.exposure_tab['Module'])):
            siaf_instrument = self.exposure_tab["Instrument"][i]
            aperture_name = self.exposure_tab['aperture'][i]
            pointing_ra = np.float(self.exposure_tab['ra'][i])
            pointing_dec = np.float(self.exposure_tab['dec'][i])
            pointing_v2 = np.float(self.exposure_tab['v2'][i])
            pointing_v3 = np.float(self.exposure_tab['v3'][i])

            # When we run across a LW grism TS aperture, save
            # the aperture name, because we'll need it when looking
            # at the accompanying SW apertuers to follow. THIS
            # RELIES ON THE LW ENTRY COMING BEFORE THE SW ENTRIES.
            if aperture_name in lw_grismts_apertures:
                lw_grismts_aperture = copy.deepcopy(aperture_name)
                lw_filter = lw_grismts_aperture.split('_')[2]
                lw_intermediate_aperture = [ap for ap in intermediate_lw_grismts_apertures if lw_filter in ap][0]

            if 'pav3' in self.exposure_tab.keys():
                pav3 = np.float(self.exposure_tab['pav3'][i])
            else:
                pav3 = np.float(self.exposure_tab['PAV3'][i])

            telescope_roll = pav3

            aperture = self.siaf[siaf_instrument][aperture_name]

            if 'NRCA5_GRISM' in aperture_name and 'WFSS' not in aperture_name:
                # This puts the source in row 29, but faster just to grab the ra, dec directly
                #local_roll, attitude_matrix, fullframesize, subarray_boundaries = \
                #    siaf_interface.get_siaf_information(self.siaf[siaf_instrument], aperture_name,
                #                                        pointing_ra, pointing_dec, telescope_roll,
                #                                        v2_arcsec=aperture.V2Ref, v3_arcsec=aperture.V3Ref)
                ra = pointing_ra
                dec = pointing_dec
            else:
                if aperture_name in sw_grismts_apertures:
                    # Special case. When looking at grism time series observation
                    # we force the pointing to be at the reference location of the
                    # LW *intermediate* aperture, rather than paying attention to
                    # the V2, V3 in the pointing file. V2, V3 from the intermediate
                    # aperture is where the source would land on the detector if
                    # the grism were not in the beam. This is exactly what we want
                    # for the SW detectors, where this is no grism.

                    # Generate an attitude matrix from this and
                    # use to get the RA, Dec in the SW apertures
                    lw_gts = self.siaf[siaf_instrument][lw_intermediate_aperture]
                    pointing_v2 = lw_gts.V2Ref
                    pointing_v3 = lw_gts.V3Ref

                local_roll, attitude_matrix, fullframesize, subarray_boundaries = \
                    siaf_interface.get_siaf_information(self.siaf[siaf_instrument], aperture_name,
                                                        pointing_ra, pointing_dec, telescope_roll,
                                                        v2_arcsec=pointing_v2, v3_arcsec=pointing_v3)

                # Calculate RA, Dec of reference location for the detector
                # Add in any offsets from the pointing file in the BaseX, BaseY columns
                ra, dec = rotations.pointing(attitude_matrix, aperture.V2Ref, aperture.V3Ref)
            aperture_ra.append(ra)
            aperture_dec.append(dec)

        self.exposure_tab['ra_ref'] = aperture_ra
        self.exposure_tab['dec_ref'] = aperture_dec

    def add_options(self, parser=None, usage=None):
        if parser is None:
            parser = argparse.ArgumentParser(usage=usage, description='Simulate JWST ramp')
        parser.add_argument("input_xml", help='XML file from APT describing the observations.')
        parser.add_argument("pointing_file", help='Pointing file from APT describing observations.')
        parser.add_argument("--output_csv", help="Name of output file containing list of observations.",
                            default=None)
        parser.add_argument("--observation_list_file", help=('Ascii file containing a list of '
                                                             'observations, times, and roll angles, '
                                                             'catalogs'), default=None)
        return parser


def get_entry(dict, entry_number):
    """Return a numbered entry from a dictionary that corresponds to the observataion_list.yaml.

    Parameters
    ----------
    dict
    entry_number

    Returns
    -------

    """
    entry_key = 'EntryNumber{}'.format(entry_number)
    for key, observation in dict.items():
        if entry_key in observation.keys():
            return observation[entry_key]


def get_filters(pointing_info):
    """Return a dictionary of instruments and filters contained within a
    pointing dictionary from an APT file.

    This function is aware that sometimes filters are installed in pupil wheels,
    and also that some wheels contain non-filter items such as the NIRCam
    weak lenses. It will return a list of those items that are indeed spectral bandpass filters.

    Parameters
    ----------
    pointing_info : dict
        This is the dictionary held in self.exposure_tab from the apt_inputs.APTInput
        class.

    Returns
    -------
    filters : dict
        Dictionary with instrument names as keys. The value for each is a
        list containing the names of the filters for that instrument present
        in the pointing dictionary.
    """
    instrument_list = set(pointing_info['Instrument'])
    filters = {}
    for inst in instrument_list:
        good = np.where(np.array(pointing_info['Instrument']) == inst.upper())[0]
        if inst.upper() == 'NIRCAM':
            short_filters = np.array(pointing_info['ShortFilter'])[good]
            long_filters = np.array(pointing_info['LongFilter'])[good]
            short_pupils = np.array(pointing_info['ShortPupil'])[good]
            long_pupils = np.array(pointing_info['LongPupil'])[good]

            short_filter_only = np.where(short_pupils == 'CLEAR')[0]
            long_filter_only = np.where(long_pupils == 'CLEAR')[0]

            filter_list = list(set(short_pupils))
            filter_list.remove('CLEAR')
            for wfsc_optic in  ['WLP8', 'WLM8', 'GDHS0', 'GDHS60']:
                if wfsc_optic in filter_list:
                    filter_list.remove(wfsc_optic)

            filter_list.extend(list(set(long_pupils)))
            filter_list.remove('CLEAR')

            filter_list.extend(list(set(short_filters[short_filter_only])))
            filter_list.extend(list(set(long_filters[long_filter_only])))

        elif inst.upper() == 'FGS':
            filter_list = ['guider1', 'guider2']

        else:
            short_filters = np.array(pointing_info['FilterWheel'])[good]
            short_pupils = np.array(pointing_info['PupilWheel'])[good]

            short_filter_only = np.where(short_pupils == 'CLEAR')[0]
            filter_list = list(set(short_pupils))
            filter_list.remove('CLEAR')
            filter_list.append(list(set(short_filters[short_filter_only])))

        filters[inst.upper()] = filter_list
    return filters


# if __name__ == '__main__':
#
#     usagestring = 'USAGE: apt_inputs.py NIRCam_obs.xml NIRCam_obs.pointing'
#
#     input = AptInput()
#     parser = input.add_options(usage=usagestring)
#     args = parser.parse_args(namespace=input)
#     input.create_input_table()



#! /usr/bin/env python

import pysiaf
from mirage.utils import siaf_interface

ra = 12.
dec = 12.
rotation = 0.
aperture = 'NRCA5_GRISM256_F444W'

siaf = pysiaf.Siaf('nircam')

roll, attitude_matrix, ffsize, subarray_bounds = siaf_interface.get_siaf_information(siaf,
                                                                                     aperture,
                                                                                     ra, dec,
                                                                                     rotation)
print(attitude_matrix)
print(roll)

loc_v2, loc_v3 = pysiaf.utils.rotations.getv2v3(attitude_matrix, ra, dec)

print(loc_v2, loc_v3)
print(siaf[aperture].V2Ref, siaf[aperture].V3Ref)

pixelx, pixely = siaf[aperture].tel_to_sci(loc_v2, loc_v3)
print(pixelx, pixely)
print(siaf[aperture].XSciRef, siaf[aperture].YSciRef)
